package ch.unibe.iam.mcast.client;

import java.rmi.RemoteException;

public class MulticastReservation {

	private static final long REFRESH_TIMEOUT = 30000l;

	private final String reservationId;

	private final Thread refreshThread;

	MulticastReservation(final String reservationId) {
		this.reservationId = reservationId;
		String threadName="renewal thread for session "+reservationId;
		this.refreshThread = new Thread(new Runnable() {

			public void run() {
				try {
					while (!Thread.interrupted()) {
						Thread.sleep(REFRESH_TIMEOUT);
						ReservationManager.refresh(reservationId);
					}
				} catch (InterruptedException e) {

				} catch (RemoteException e) {
					throw new RuntimeException(e);
				}

			}

		},threadName);
		this.refreshThread.setDaemon(true);
		
		this.refreshThread.start();
	}

	public boolean reserveResources(int bandwidth, int maxDelay, int jitter) throws ReservationSystemFailureException {
		try {
			return ReservationManager.reserve(reservationId, bandwidth, maxDelay, jitter);
		} catch (RemoteException e) {
			throw new ReservationSystemFailureException(e);
		}
	}

	public void release() throws ReservationSystemFailureException {
		this.refreshThread.interrupt();
		try {
			ReservationManager.release(reservationId);
		} catch (RemoteException e) {
			throw new ReservationSystemFailureException(e);
		}
	}

	public String getReservationId() {
		return reservationId;
	}
	
	

}
