package ch.unibe.iam.mcast.client;

import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.rmi.RemoteException;

import javax.xml.namespace.QName;
import javax.xml.rpc.ServiceException;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;

public class ReservationManager {

	private static final String endpoint;

	private static String getStringEnv(String name, String def) {
		String ret=System.getenv(name);
		if (ret == null) {
			ret=def;
		}
			
		return ret;
	
	}

	static {
		endpoint = "http://localhost:"+getStringEnv("MCAST_HTTP_PORT","8090")+"/services/EuQoSReservationService";
	}
	
	private static final Call sessionCall;

	private static final Call reserveCall;

	private static final Call releaseCall;
	
	private static final Call refreshCall;


	static String session(InetSocketAddress groupAddress, int protocol) throws RemoteException {
		return (String) sessionCall.invoke(new Object[] { groupAddress.getAddress().getHostAddress(), new Integer(groupAddress.getPort()), new Integer(protocol) });
	}

	static boolean reserve(String sessionId, int bw, int maxDelay, int jitter) throws RemoteException {
		return ((Boolean) reserveCall.invoke(new Object[] { sessionId, new Integer(bw), new Integer(maxDelay), new Integer(jitter) })).booleanValue();
	}
	
	static void release(String sessionId) throws RemoteException {
		releaseCall.invoke(new Object[] { sessionId});
	}
	
	static void refresh(String sessionId) throws RemoteException {
		refreshCall.invoke(new Object[] { sessionId});
	}
	
	public static MulticastReservation createSession(InetSocketAddress multicastDestination, int protocol) throws ReservationSystemFailureException {
		
		String sessionId;
		try {
			sessionId = session(multicastDestination,protocol);
		} catch (RemoteException e) {
			throw new ReservationSystemFailureException(e);
		}
		MulticastReservation reservation=new MulticastReservation(sessionId);
		return reservation;
	}

	static {

		sessionCall = createCall(endpoint, "session");
		reserveCall = createCall(endpoint, "reserve");
		releaseCall = createCall(endpoint, "release");
		refreshCall = createCall(endpoint, "refresh");
	}

	private static Call createCall(String endpoint, String operationName) {
		Service service = new Service();
		Call call;
		try {
			call = (Call) service.createCall();
		} catch (ServiceException e) {
			throw new RuntimeException(e);
		}

		try {
			call.setTargetEndpointAddress(new java.net.URL(endpoint));
		} catch (MalformedURLException e) {
			throw new RuntimeException(e);
		}

		call.setOperationName(new QName(operationName));
		return call;
	}

	
	
	public static void main(String argv[]) throws ReservationSystemFailureException, InterruptedException {
		MulticastReservation mr=createSession(new InetSocketAddress("239.1.2.3",1245), 0);
		mr.reserveResources(2*1024, 200, 20);
		Thread.sleep(50000);
		mr.release();
		Thread.sleep(2000);
	}
}
