package ch.unibe.iam.mcast.hivemind.component;

import java.net.InetSocketAddress;
import java.net.URL;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hivemind.events.RegistryShutdownListener;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.webapp.WebAppContext;

import ch.unibe.iam.mcast.core.MulticastRoutingTable;
import ch.unibe.iam.mcast.hivemind.component.iface.MulticastRouter;
import ch.unibe.iam.mcast.hivemind.component.iface.WebInterface;

public class RouterWebInterface implements WebInterface, RegistryShutdownListener {

	private static final String WEBDEFAULT_XML = "webdefault.xml";

	private static final String WEBAPP_PATH = "/router-controller.war";

	private static final Log log = LogFactory.getLog(RouterWebInterface.class);

	private final InetSocketAddress bindAddress = new InetSocketAddress(Integer.parseInt(ComponentUtil.getStringEnv("MCAST_HTTP_PORT","8090")));

	private Server server = new Server();

	private Connector connector = new SelectChannelConnector();

	private final MulticastRoutingTable routingTable;

	private final MulticastRouter router;

	public RouterWebInterface(MulticastRouter router) {
		this.router = router;
		// this.bindAddress = bindAddress;
		this.routingTable = router.getMulitcastRoutingTable();
		if (bindAddress.getHostName() != null) {
			connector.setHost(bindAddress.getHostName());
		}
		connector.setPort(bindAddress.getPort());
		server.setConnectors(new Connector[] { connector });
	}

	public void start() throws Exception {
		URL webappURL = getClass().getResource(WEBAPP_PATH);
		if (webappURL != null) {
			WebAppContext routerWac = new WebAppContext();
			URL defaultDescriptorURL = this.getClass().getResource(WEBDEFAULT_XML);
			routerWac.setDefaultsDescriptor(defaultDescriptorURL.toExternalForm());
			routerWac.setAttribute("multicast.routing.table", routingTable);
			routerWac.setAttribute("multicast.router", router);
			routerWac.setContextPath("/");
			routerWac.setWar(webappURL.toString());
			server.setHandler(routerWac);
		} else {
			log.error("could not find webapp application with path " + WEBAPP_PATH);
		}

		server.start();
	}

	public void stop() {

		try {
			server.stop();
		} catch (Exception e) {
			log.warn(e);
		}
	}

	public String toString() {
		return "WebappContainer(" + bindAddress + ")";
	}

	public void registryDidShutdown() {

		stop();

	}

}
