package ch.unibe.iam.mcast.hivemind.component.euqos;

import java.net.InetSocketAddress;
import java.util.Collections;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.euqos.client.qcm.QCM;
import org.euqos.client.qcm.QCMInitializationException;
import org.euqos.client.qcm.bqcm.BQCM;
import org.euqos.client.qcm.bqcm.BQCMException;
import org.euqos.client.qcm.bqcm.IBQCM;
import org.euqos.client.qcm.bqcm.NotInSessionException;
import org.euqos.client.qcm.bqcm.NotRegisteredException;
import org.euqos.common.DataStream;
import org.euqos.common.QoSAnswer;
import org.euqos.common.QoSRequest;
import org.euqos.common.QosCharacteristics;

import ch.unibe.iam.mcast.hivemind.component.ComponentUtil;
import ch.unibe.iam.mcast.router.RouterUtil;
import ch.unibe.iam.mcast.router.p2p.ConnectionInformation;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;
import ch.unibe.iam.mcast.router.qos.QoSReservationService;
import ch.unibe.iam.mcast.router.qos.QosReservationException;
import ch.unibe.iam.mcast.router.qos.ReservedResources;

public class QCMComponent implements QoSReservationService {

	private static final String ENV_EUQOS_SEND_STRENGTH = "EUQOS_SEND_STRENGTH";

	private static final String ENV_EUQOS_RECV_STRENGTH = "EUQOS_RECV_STRENGTH";

	private static final String ENV_EUQOS_TRANSPORT_PROTOCOL = "EUQOS_TRANSPORT_PROTOCOL";

	private static final String ENV_EUQOS_APPLICATION_TYPE = "EUQOS_APPLICATION_TYPE";

	private static final Log log = LogFactory.getLog(QCMComponent.class);

	// O

	// private final Map<QoSRequirement, QosClassDescriptor> reqClasses = new
	// HashMap<QoSRequirement, QosClassDescriptor>();
	//
	// {
	// reqClasses.private List appMediaList;
	//	
	// {
	// try {
	// appMediaList=AppParamsReader.readAppMediaDescriptionList();
	// } catch (IOException e) {
	// throw new RuntimeException(e);
	// }
	// }put(QoSRequirement.HIGH, new
	// QosClassDescriptor(getIntEnv(ENV_BANDWIDTH,1023),
	// getIntEnv(ENV_DELAY,300), getIntEnv(ENV_JITTER,1)));
	// reqClasses.put(QoSRequirement.MEDIUM, new QosClassDescriptor(1024, 400,
	// 1));
	// }

	// private final Map<InetSocketAddress, EuQoSReservedResources>
	// reservedResourcesByConnection = new HashMap<InetSocketAddress,
	// EuQoSReservedResources>();

	// private final Map<String, EuQoSReservedResources>
	// reservedResourcesBySessionId = new HashMap<String,
	// EuQoSReservedResources>();

	private final String uname;

	private final String password;

	private IBQCM qcm;

	private static final String appType = ComponentUtil.getStringEnv(ENV_EUQOS_APPLICATION_TYPE, "MMStreaming");

	public QCMComponent(String uname, String password) {
		this.uname = uname;
		this.password = password;
		try {
			// this.qcm = new BQCM();
			this.qcm = QCM.getOnDemandBQCM();
		} catch (QCMInitializationException e) {
			throw new RuntimeException(e);
		}
	}

	public void stop() {

	}

	public String makeReservation(InetSocketAddress from, InetSocketAddress to, double bw, double delay, double jitter) throws ReservationFailedException {
				
		QoSRequest req = new QoSRequest();

		req.setApplicationType(appType);
		DataStream ds = new DataStream();
		ds.setDestinationPort(to.getPort());
		ds.setSourcePort(from.getPort());
		QosCharacteristics qc = new QosCharacteristics();
		qc.setMaxDelay((int) delay);
		qc.setMaxJitter((int) jitter);
		qc.setMaxLoss(0.005f);
		qc.setPeakBitRate((int) (bw));

		ds.setQosCharacteritics(Collections.singletonList(qc));
		// md.setReservationDirection(getStringEnv(ENV_EUQOS_RESERVATION_DIRECTION,
		// "sendrcv"));
		ds.setRecvStrength(ComponentUtil.getStringEnv(ENV_EUQOS_RECV_STRENGTH, "none"));
		ds.setSendStrength(ComponentUtil.getStringEnv(ENV_EUQOS_SEND_STRENGTH, "mandatory"));
		ds.setTransportProtocol(ComponentUtil.getStringEnv(ENV_EUQOS_TRANSPORT_PROTOCOL, "TCP"));

		req.setDataStreamList(Collections.singletonList(ds));

		req.setDestinationIP(to.getAddress().getHostAddress());
		req.setSourceIP(from.getAddress().getHostAddress());
		req.setUserId(uname);
		req.setPassword(password);

		try {
			QoSAnswer ans = qcm.performReservation(req);
			if (ans == null) {
				throw new ReservationFailedException("could not perform reservation");

			}
			return ans.getSessionId();
		} catch (BQCMException e1) {
			log.warn(e1);
			throw new ReservationFailedException(e1);
		} catch (Throwable t) {
			log.warn(t);
			if (t instanceof Error) {
				throw (Error) t;
			} else if (t instanceof RuntimeException) {
				throw (RuntimeException) t;
			} else {
				throw new RuntimeException(t);
			}
		}
	}

	// public List getAppMediaDescriptionList() {
	// return appMediaList;
	// }
	//
	// public void notifyDeregistrationResult(boolean arg0) {
	//
	// }
	//
	// public void notifyRegistrationResult(RegistrationResult
	// registrationResult) {
	//
	// // appFrame.setRegisteredStatus(registrationResult
	// // .getRegistrationFlag(), regDialog.getUserId());
	// if (!registrationResult.getRegistrationFlag()) {
	// // log("Registration failed !\r\nReason: "
	// // + registrationResult.getMessage());
	// // showErrorMessage("Registration failed !\r\nReason: "
	// // + registrationResult.getMessage());
	//
	// // registered = false;
	// // doRegister();
	// } else {
	// // appFrame.log("Registration successful");
	// // registered = true;
	// registrationId = registrationResult.getRegistrationId();
	// QoSProfile[] profiles = QCM.getXQCM().askQoSProfiles(registrationId);
	// for (QoSProfile a : profiles) {
	// System.out.println(a);
	// }
	// hasRegistered();
	// }
	//
	// }
	//
	// public EventResult notifySessionEvent(SessionEvent evt) {
	//
	// System.out.println("event: " + evt);
	//
	// EventResult ret = null;
	// if (evt instanceof SessionEstablishResultEvent) {
	// SessionEstablishResultEvent sere = (SessionEstablishResultEvent) evt;
	// String sessionId = sere.getSessionId();
	// // EuQoSReservedResources reserved;
	// // synchronized (reservedResourcesBySessionId) {
	// // reserved = reservedResourcesBySessionId.remove(sessionId);
	// // }
	// //
	// // if (reserved != null) {
	// //// synchronized (reservedResourcesByConnection) {
	// ////
	// reservedResourcesByConnection.remove(reserved.getConnectionInformation().getLocalAddress());
	// //// }
	// // if (!sere.isSetupFlag()) {
	// // reserved.setException(new QosReservationException(sere.getMessage()));
	// // }
	// // reserved.unblock();
	// // }
	// } else if (evt instanceof SessionCloseResultEvent) {
	// SessionCloseResultEvent scre = (SessionCloseResultEvent) evt;
	// if (log.isDebugEnabled()) {
	// log.debug("session " + scre.getSessionId() + " closed. Message: " +
	// scre.getMessage());
	// }
	// } else if (evt instanceof SessionRingingEvent) {
	//
	// SessionRingingEvent ringing = (SessionRingingEvent) evt;
	// //this.appMediaList=ringing.getConnectionCharacteristics().getMediaList();
	// //System.out.println("app media list: "+this.appMediaList);
	// SessionRingingResult rslt = new SessionRingingResult();
	//
	// try {
	// rslt.setSourceIP(InetAddress.getLocalHost().getHostAddress());
	// } catch (UnknownHostException e) {
	// throw new RuntimeException(e);
	// }
	// rslt.setReponseFlag(true);
	// if (log.isDebugEnabled()) {
	// log.debug("ringing from " + ringing.getCallerLogin());
	// }
	// ret = rslt;
	// } else if (evt instanceof SessionProgressEvent) {
	// SessionProgressEvent spe = (SessionProgressEvent) evt;
	// // ConnectionInformation ci =
	// extractConnInfo(spe.getConnectionCharacteristics());
	// // EuQoSReservedResources res;
	// // synchronized (reservedResourcesByConnection) {
	// // res = reservedResourcesByConnection.get(ci.getLocalAddress());
	// // if (res != null) {
	// // res.setSessionId(spe.getSessionId());
	// // }
	// // }
	// // synchronized (reservedResourcesBySessionId) {
	// // reservedResourcesBySessionId.put(spe.getSessionId(), res);
	// // }
	//
	// }
	// // TODO Auto-generated method stub
	// return ret;
	// }
	//
	// // private ConnectionInformation
	// extractConnInfo(ConnectionCharacteristics cc) {
	// // MediaDescription md = (MediaDescription) cc.getMediaList().get(0);
	// // ConnectionInformation ret = new ConnectionInformation(new
	// InetSocketAddress(cc.getSourceIP(), md.getSourcePort()), new
	// InetSocketAddress(cc.getDestinationIP(), md
	// // .getDestinationPort()));
	// // return ret;
	// // }

	public ReservedResources reserveResources(InetSocketAddress from, InetSocketAddress to, QoSRequirement reqirement) throws QosReservationException {
		QosClassDescriptor cls;

		cls = new QosClassDescriptor(reqirement.getBandwidth() / 1000.0, reqirement.getDelay(), reqirement.getJitter());

		ConnectionInformation connectionInformation = new ConnectionInformation(from, to);

		boolean isBE = reqirement.isBestEffort();

		System.out.println("changing qos from " + from + " to " + to + " req: " + reqirement + " BE: " + isBE);

		if (!isBE) {
			try {

				System.out.println("******* Making reservation " + from + " -> " + to);

				String reservationId = makeReservation(from, to, cls.getBandwidth(), cls.getDelay(), cls.getJitter());
				EuQoSReservedResources reserved = new EuQoSReservedResources(connectionInformation, reservationId);
				return reserved;

			} catch (ReservationFailedException e) {
				throw new QosReservationException(e);
			}
		} else {
			return new ReservedResources() {

				public void releaseResources() {

				}

			};
		}

	}

	private class EuQoSReservedResources implements ReservedResources {

		private final String sessionId;

		// private QosReservationException exception;

		private final ConnectionInformation connectionInformation;

		public EuQoSReservedResources(ConnectionInformation connectionInformation, String reservationId) {
			this.connectionInformation = connectionInformation;
			this.sessionId = reservationId;

		}

		public String getSessionId() {

			return sessionId;
		}

		public void releaseResources() {
			try {
				System.out.println("releasing " + sessionId);
				qcm.closeReservation(sessionId, appType);
			} catch (NotRegisteredException e) {
				log.error(e);
			} catch (NotInSessionException e) {
				log.error(e);
			} catch (BQCMException e) {
				log.error(e);
			}
		}

	}

}
