package ch.unibe.iam.mcast.core;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import ch.unibe.iam.mcast.net.ip.IPV4Packet;

public abstract class LocalNetworkInterface implements MulticastRoutingDestination {

    private final String name;

    private final InetAddress address;

    private final ArrayList[] v4PacketListeners = new ArrayList[256];

   // private final Map ethernetPacketListeners = new HashMap();

    private final ArrayList localNetworkInterfaceEventListener = new ArrayList();

    public LocalNetworkInterface(String name, InetAddress address) {
        if (name == null) {
            throw new IllegalArgumentException("name should not be null");
        }
        this.name = name;

        if (address == null) {
            throw new IllegalArgumentException("address should not be null");
        }
        this.address = address;
    }

    public String getName() {
        return name;
    }

    public int hashCode() {
        return name.hashCode() ^ address.hashCode();

    }

    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof LocalNetworkInterface)) {
            return false;
        }
        LocalNetworkInterface other = (LocalNetworkInterface) o;
        return other.name.equals(this.name) && other.address.equals(address);
    }

    public void addIPV4PacketListener(byte ipProtocol, IPV4PacketListener listener) {
        if (listener == null) {
            throw new IllegalArgumentException("listener should not be null");
        }
        ArrayList listenerList = v4PacketListeners[ipProtocol & 0xff];
        if (listenerList == null) {
            listenerList = new ArrayList();
            v4PacketListeners[ipProtocol & 0xff] = listenerList;
        }
        if (!listenerList.contains(listener)) {
            listenerList.add(listener);
        }
    }

    public void removeIPV4PacketListener(byte ipProtocol, IPV4PacketListener listener) {
        if (listener == null) {
            throw new IllegalArgumentException("listener should not be null");
        }
        v4PacketListeners[ipProtocol & 0xff].remove(listener);

    }

    protected final void onIPV4Packet(IPV4Packet packet) {
        if (packet == null) {
            throw new IllegalArgumentException("packet should not be null");
        }
        int protocol = packet.getProtocol() & 0xff;
        if (v4PacketListeners[protocol] != null) {
            for (int i = 0; i < v4PacketListeners[protocol].size(); i++) {
                ((IPV4PacketListener) v4PacketListeners[protocol].get(i)).onIPPacket(this,packet);
            }
        }
    }
    
    protected final boolean hasCookedListener(byte protocol) {
    	int proto = protocol & 0xff;
    	return v4PacketListeners[proto] != null;
    }

    public void addLocalNetworkInterfaceEventListener(LocalNetworkInterfaceEventListener listener) {
        if (listener == null) {
            throw new IllegalArgumentException("listener should not be null");
        }
        if (!localNetworkInterfaceEventListener.contains(listener)) {
            localNetworkInterfaceEventListener.add(listener);
        }

    }

    public void removeLocalNetworkInterfaceEventListener(LocalNetworkInterfaceEventListener listener) {
        if (listener == null) {
            throw new IllegalArgumentException("listener should not be null");
        }
        localNetworkInterfaceEventListener.remove(listener);
    }

    protected void onLocalNetworkInterfaceClosed() {
        for (int i = 0; i < localNetworkInterfaceEventListener.size(); i++) {
            ((LocalNetworkInterfaceEventListener) localNetworkInterfaceEventListener.get(i)).onLocalNetworkInterfaceShutdown(this);
        }
    }

    public abstract void close();
}
