package ch.unibe.iam.mcast.core;

import java.net.InetSocketAddress;
import java.util.ArrayList;

import ch.unibe.iam.mcast.router.p2p.ConnectionInformation;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;

public abstract class PeerConnection implements MulticastRoutingDestination {

    //private final InetSocketAddress peerAddress;
    private final ArrayList<PeerConnectionListener> listeners=new ArrayList<PeerConnectionListener>();
	//private final InetSocketAddress localAddress;
    private final ConnectionInformation connectionInformation;
    

    public PeerConnection(InetSocketAddress localAddress,InetSocketAddress peerAddress) {

		if (localAddress == null) {
            throw new IllegalArgumentException("localAddress should not be null");
        }
       
        
		if (peerAddress == null) {
            throw new IllegalArgumentException("peerAddress should not be null");
        }
		this.connectionInformation=new ConnectionInformation(localAddress,peerAddress);
		
//		this.localAddress = localAddress;
//        this.peerAddress = peerAddress;
    }

    public int hashCode() {
        return connectionInformation.hashCode();

    }

    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        
        if (!(o instanceof PeerConnection)) {
            return false;
        }
        
        PeerConnection other = (PeerConnection) o;
        return this.connectionInformation.equals(other.connectionInformation);
    }
    
    public void addPeerMessageListener(PeerConnectionListener listener) {
        if (listener == null) {
            throw new IllegalArgumentException("listener should not be null");
        }
        
        if (!listeners.contains(listener)) {
            listeners.add(listener);
        }
    }
    
    public void removePeerMessageListener(PeerConnectionListener listener) {
        listeners.remove(listener);
    }
    
    protected void onNewMessage(P2PMessage msg) {
        if (msg == null) {
            throw new IllegalArgumentException("msg should not be null");
        }
        for (int i=0; i<listeners.size(); i++) {
            listeners.get(i).onPeerMessage(this,msg);
        }
    }
    
    protected void onConnectionClosed() {
    	boolean isInterrupted=Thread.interrupted();
        for (int i=0; i<listeners.size(); i++) {
            listeners.get(i).onPeerConnectionClosed(this);
        }
        if (isInterrupted) {
        	Thread.currentThread().interrupt();
        }
    }
    
    public abstract void sendMessage(P2PMessage message);

    public String toString() {
        return ("PeerConnection("+connectionInformation+")");
    }
    
    public abstract void close();

    public ConnectionInformation getConnectionInformation() {
        return connectionInformation;
    }    

    
}
