package ch.unibe.iam.mcast.core.java;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.PeerConnection;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.util.ByteBufferUtil;
import ch.unibe.iam.mcast.router.p2p.message.MalformedP2PMessageException;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessageConstants;
import ch.unibe.iam.mcast.router.p2p.message.P2PObjectConstants;
import ch.unibe.iam.mcast.router.p2p.message.P2PObjectUtil;
import ch.unibe.iam.mcast.router.p2p.message.data.IPV4DataPacketMessage;

public class JavaPeerConnection extends PeerConnection implements JavaMulticastRoutingDestination {

	private static final int INITIAL_CAPACITY = 8192;

	private static final int LOW_WATER_MARK = 2048;

	private static final Log log = LogFactory.getLog(JavaPeerConnection.class);

//	private final SocketChannel channel;

	private final PrioritySendQueue sendQueue = new PrioritySendQueue();

	private ByteBuffer writingBuffer = ByteBuffer.allocate(INITIAL_CAPACITY);

	private ByteBuffer messageCreationBuffer = ByteBuffer.allocate(INITIAL_CAPACITY);

	private boolean running;

	private final SendingThread sending;

	private final ReadingThread reading;

	private final ArrayList dataMessageListeners = new ArrayList();

	public JavaPeerConnection(SocketChannel channel) {

		super((InetSocketAddress) channel.socket().getLocalSocketAddress(), (InetSocketAddress) channel.socket().getRemoteSocketAddress());
		
		try {
			channel.socket().setSoLinger(false, 0);
			channel.socket().setTcpNoDelay(true);
		} catch (SocketException e) {
			log.fatal(e);
			throw new RuntimeException(e);
		}
		if (channel == null) {
			throw new IllegalArgumentException("channel should not be null");
		}
		
		sending = new SendingThread(channel);
		reading = new ReadingThread(sending,channel);
		
		
//		this.channel = channel;
	}

	public synchronized void start() {
		if (running) {
			throw new IllegalStateException("already running");
		}
		sending.start();
		reading.start();
		running = true;
	}

	public synchronized void stop() {
		if (!running) {
			throw new IllegalStateException("not running");
		}

		reading.interrupt();
		// try {
		// reading.join();
		// } catch (InterruptedException e) {
		// log.warn(e);
		// Thread.currentThread().interrupt();
		// }

	}

	public void sendMessage(P2PMessage message) {
		sendQueue.add(0, message);
	}

	public void sendMessage(P2PMessage message, int priority) {
		sendQueue.add(priority, message);
	}

	private class SendingThread extends Thread {

		private final SocketChannel channel;

		public SendingThread(SocketChannel channel) {
			this.channel = channel;
			
		}

		public void run() {
			// System.out.println("seding thread running");
			while (!interrupted()) {
				try {
					writingBuffer.clear();

					sendQueue.fillBuffer(writingBuffer, LOW_WATER_MARK);
					// System.out.println("writing "+writingBuffer);

					// P2PMessage message = sendQueue.next();
					//
					// int size = message.getMessageLength();
					//                    
					// if (writingBuffer.capacity() < size) {
					// writingBuffer = ByteBuffer.allocate(size);
					// }
					//
					// // System.out.println("sending message "+message);
					//
					// writingBuffer.clear();
					// writingBuffer.limit(size);
					// message.serialize(writingBuffer);
					// writingBuffer.flip();
					// // System.out.println("+++++ sending:");
					// // ByteBufferUtil.dumpByteBuffer(writingBuffer);
					writingBuffer.flip();
					// System.out.println("writing "+writingBuffer);
					while (writingBuffer.remaining() > 0) {
						channel.write(writingBuffer);
					}
					// System.out.println("after writing "+writingBuffer);

					// System.out.println("---- sent!");

				} catch (InterruptedException e) {
					// ignore
					return;
				} catch (IOException e) {
					log.warn(e);
					// also ignore
					return;
				}
			}
			// System.out.println("seding thread stopped");

		}
	}

	private class ReadingThread extends Thread {

		private final SendingThread sendingThread;
		private final SocketChannel channel;

		public ReadingThread(SendingThread sendingThread, SocketChannel channel) {
			super("Reading thread "+channel.socket().getLocalAddress());
			this.sendingThread = sendingThread;
			this.channel=channel;
		}

		public void run() {
			ByteBuffer readBuffer = ByteBuffer.allocate(INITIAL_CAPACITY);
			readBuffer.limit(4);
			try {
				while (!interrupted()) {
					while (readBuffer.remaining() > 0) {
						int read = channel.read(readBuffer);
						if (read < 0) {
							return;
						}

					}
					int length = readBuffer.getShort(2) & 0xffff;
					// System.out.println("++++length: "+length);
					if (readBuffer.position() == length) {
						readBuffer.flip();
						// byte[] data = new byte[readBuffer.remaining()];
						// readBuffer.get(data);

						// intercept the data objects
						// int oldPosition=readBuffer.position();
						// try {
						// P2PMessage message = P2PMessage
						// .deserializeMessage(readBuffer);
						// System.out.println("got message: "+message);
						// //onNewMessage(message);
						// } catch (MalformedP2PMessageException e) {
						// log.warn("communication error", e);
						// throw new IOException("protocol error");
						// }
						// readBuffer.position(oldPosition);

						if (isDataMessage(readBuffer)) {
							int originalPosition = readBuffer.position();
							int originalLimit = readBuffer.limit();
							for (int i = 0; i < dataMessageListeners.size(); i++) {
								DataMessageListener listener = (DataMessageListener) dataMessageListeners.get(i);
								readBuffer.position(originalPosition + 8);
								readBuffer.limit(originalLimit);
								// System.out.println("buffer: "+readBuffer);
								listener.onNewDataMessage(JavaPeerConnection.this, readBuffer);

							}
							readBuffer.position(originalPosition);
							readBuffer.limit(originalLimit);
						} else {

							//

							P2PMessage message;
							try {
								message = P2PMessage.deserializeMessage(readBuffer);
								// System.out.println("got message: "+message);
								onNewMessage(message);
							} catch (MalformedP2PMessageException e) {
								log.warn("communication error", e);
								throw new IOException("protocol error");
							}

						}
						readBuffer.clear();
						readBuffer.limit(4);
					} else {
						if (readBuffer.capacity() < length) {
							ByteBuffer newBuffer = ByteBuffer.allocate(length);
							readBuffer.flip();
							newBuffer.put(readBuffer);
							readBuffer = newBuffer;
						}
						readBuffer.limit(length);
					}
				}
			} catch (IOException e) {
				if (log.isDebugEnabled()) {
					log.debug(e);
				}
			} finally {

				onConnectionClosed();
				try {
					channel.close();
				} catch (IOException e1) {
//					if (log.isWarnEnabled()) {
//						log.warn(e1);
//					}
				}
				sendingThread.interrupt();
				try {
					channel.close();
				} catch (IOException e) {
					// just 4get it!
				}
			}

		}

		private boolean isDataMessage(ByteBuffer readBuffer) {
			short type = readBuffer.getShort(readBuffer.position());
			return type == P2PMessageConstants.IPV4_DATA_PACKET_MESSAGE;
		}

	}

	public void close() {
		stop();

	}

	public void sendIPV4Packet(IPV4Packet packet, int priority) {
		IPV4DataPacketMessage msg = new IPV4DataPacketMessage(packet);
		sendMessage(msg, priority);
	}

	public synchronized void sendRawIPV4Packet(ByteBuffer buffer, int priority) {
		messageCreationBuffer.clear();
		messageCreationBuffer.putShort(P2PMessageConstants.IPV4_DATA_PACKET_MESSAGE);
		int length = buffer.remaining();
		messageCreationBuffer.putShort((short) (length + 8));
		P2PObjectUtil.writeObjectHeader(messageCreationBuffer, P2PObjectConstants.CLASS_DATA, P2PObjectConstants.TYPE_IP_PACKET_DATA, length);
		messageCreationBuffer.put(buffer);
		messageCreationBuffer.flip();
		sendQueue.add(priority, messageCreationBuffer);
	}

	public synchronized void addDataMessageListener(DataMessageListener listener) {
		this.dataMessageListeners.add(listener);
	}

	public synchronized void removeDataMessageListener(DataMessageListener listener) {
		this.dataMessageListeners.remove(listener);
	}


}
