/*
 * Created on Dec 3, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package ch.unibe.iam.mcast.net.ethernet;

import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.net.Packet;
import ch.unibe.iam.mcast.net.arp.ARPPacket;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.MalformedPacketException;
import ch.unibe.iam.mcast.net.ip.util.ByteBufferUtil;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window - Preferences - Java - Code Style - Code Templates
 */
public class EthernetPacket implements Packet {

    private final EthernetAddress destination;

    private final EthernetAddress source;

    private final short protocol;

    private final byte[] payload;

    public byte[] getPayload() {
        return payload;
    }

    public EthernetAddress getDestination() {
        return destination;
    }

    public short getProtocol() {
        return protocol;
    }

    public EthernetAddress getSource() {
        return source;
    }

    private final Packet transportedPacket;

    public EthernetPacket(EthernetAddress destinationAddress, EthernetAddress sourceAddress, IPV4Packet packet) {
        this.protocol = EthernetConstants.ETH_P_IP;
        this.source = sourceAddress;
        this.destination = destinationAddress;
        this.payload = null;
        this.transportedPacket = packet;

    }

    /**
     *  
     */
    private EthernetPacket(EthernetAddress destinationAddress, EthernetAddress sourceAddress, short protocol, byte[] data) throws MalformedPacketException {

        this.destination = destinationAddress;
        this.source = sourceAddress;

        this.protocol = protocol;
        this.payload = data;

        ByteBuffer buffer = ByteBuffer.wrap(data);

        //System.out.println("protocol: ")

        switch (protocol) {
        case EthernetConstants.ETH_P_ARP:

            this.transportedPacket = ARPPacket.parseARPPacket(buffer);
            break;
        case EthernetConstants.ETH_P_IP:
            Packet pck = null;

            pck = IPV4Packet.createIPV4Packet(buffer);

            this.transportedPacket = pck;
            break;
        default:
            this.transportedPacket = null;
        }

    }

    public static EthernetPacket createPacket(ByteBuffer buffer) throws MalformedPacketException {
        if (buffer.remaining() < 14) {
            throw new MalformedPacketException("buffer shorter than ethernet frame");
        }
        EthernetAddress destination = new EthernetAddress(buffer);
        EthernetAddress source = new EthernetAddress(buffer);        
        short protocol = buffer.getShort();
        byte[] data = new byte[buffer.remaining()];
        buffer.get(data);

        return new EthernetPacket(source, destination, protocol, data);

    }

    public String toString() {
        StringBuffer string = new StringBuffer("Ethernet Packet(");
        string.append("from: ");
        string.append(source.toString());
        string.append(" to: ");
        string.append(destination.toString());
        string.append(" protocol: ");
        string.append(protocol & 0xffff);
        string.append(" transported packet: ");
        string.append(transportedPacket);
        string.append(" data: ");
        ByteBuffer buf = ByteBuffer.wrap(payload);
        string.append(ByteBufferUtil.byteBufferToString(buf));
        string.append(")");

        return string.toString();

    }

    public Packet getTransportedPacket() {

        return transportedPacket;
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.net.Packet#serializePacket(java.nio.ByteBuffer)
     */
    public void serializePacket(ByteBuffer buffer) {
        destination.serializeAddress(buffer);
        source.serializeAddress(buffer);
        buffer.putShort(protocol);
        if (payload != null) {
            buffer.put(payload);
        } else {
            transportedPacket.serializePacket(buffer);
        }

    }

    private int getPayloadLength() {
        if (payload != null) {
            return payload.length;
        } else {
            return transportedPacket.getPacketLength();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.net.Packet#getPacketLength()
     */
    public int getPacketLength() {

        return EthernetConstants.ETHERNET_ADDRESS_LENGTH * 2 + 2 + getPayloadLength();
    }

}