/*
 * Created on Oct 8, 2004
 *
 */
package ch.unibe.iam.mcast.net.ip;

import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.net.ip.util.HashUtil;

/**
 * @author dragan
 *  
 */
public class UnknownIPV4Packet extends IPV4Packet {

    private byte[] data;

    /**
     * @param buffer
     * @throws MalformedPacketException
     */
    public UnknownIPV4Packet(IPV4PacketHeader header, ByteBuffer buffer) throws MalformedPacketException {
        super(header);
        this.data = new byte[buffer.remaining()];
        buffer.get(data);
    }

    public UnknownIPV4Packet(byte protocol, byte[] data) {
        super(protocol);
        setData(data);
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.ip.IPV4Packet#serializeData(java.nio.ByteBuffer)
     */
    protected void serializeData(ByteBuffer buffer) {
        buffer.put(data);

    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.ip.IPV4Packet#getDataLengt()
     */
    protected int getDataLengt() {
        return data.length;
    }

    /**
     * @return Returns the data.
     */
    public byte[] getData() {
        return data;
    }

    /**
     * @param data
     *            The data to set.
     */
    public void setData(byte[] data) {
        if (data == null) {
            throw new IllegalArgumentException("data should not be null");
        }
        this.data = data;
    }

    public void setProtocol(byte protocol) {
        this.header.setProtocol(protocol);
    }
    

    public boolean equals(Object obj) {
        if (!(obj instanceof UnknownIPV4Packet)) {
            return false;
        } else {
            UnknownIPV4Packet other = (UnknownIPV4Packet) obj;
            return this.header.equals(other.header) && HashUtil.areEqual(this.data, other.data);
        }
    }
    
    
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return header.hashCode() ^ HashUtil.getHash(data);
	}
	
	
	
	
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return "UnknownIPV4Packet(protocol: "+(short)(getProtocol() & 0xff)+" from: "+getSourceAddress()+" to: "+getDestinationAddress()+" ttl: "+getTimeToLive()+")";
    }
}