/*
 * Created on Oct 8, 2004
 *
 */
package ch.unibe.iam.mcast.net.ip.igmp;

import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.IPV4PacketHeader;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.net.ip.util.InternetChecksumUtil;

/**
 * @author dragan
 *  
 */
public abstract class IGMPPacket extends IPV4Packet {

	private static final int CHECKSUM_OFFSET = 2;

	public IGMPPacket() {
		super(IPProtocolConstants.IGMP);
		super.setTimeToLive((byte)1); // has to be one for IGMP messages
	}

	public IGMPPacket(IPV4PacketHeader header) {
		super(header);
		super.setTimeToLive((byte)1); // has to be one for IGMP messages
	}

	final protected void serializeData(ByteBuffer buffer) {
		int startPosition = buffer.position();

		int checksumPosition = startPosition + CHECKSUM_OFFSET;
		serializeDataInternal(buffer);
		int newPosition = buffer.position();
		int oldLimit=buffer.limit();
		buffer.limit(buffer.position());
		
		buffer.position(startPosition);
		short checksum = InternetChecksumUtil.computeInetChecksum(buffer);
		buffer.position(newPosition);
		buffer.limit(oldLimit);
		buffer.putShort(checksumPosition, checksum);
	}
	
	final public void setTimeToLive(byte b) {
	  throw new RuntimeException("Not allowed");
	}

	/**
	 * @param buffer
	 */
	protected abstract void serializeDataInternal(ByteBuffer buffer);

}