/*
 * Created on Oct 18, 2004
 */
package ch.unibe.iam.mcast.net.ip.igmp;

import java.net.Inet4Address;
import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.net.ip.IPV4PacketHeader;
import ch.unibe.iam.mcast.net.ip.util.IPV4AddressUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * @author brogle
 *
 * Represents an IGMP host membership report message
 */
public class IGMPV2LeaveGroupPacket extends IGMPPacket {
	
	/**
	 * Logging 
	 */
	private static Log log = LogFactory.getLog(IGMPV2LeaveGroupPacket.class);
	
	// private byte maxResponseTime; // IGMPv2
	
	/**
	 * Default multicast group address to use
	 */
	private Inet4Address groupAddress=IPV4AddressUtil.NULL_IPV4ADDRESS;
	
	/**
	 * Dummy constructor 
	 */
	public IGMPV2LeaveGroupPacket() {
		super();
	}
	
	/**
	 * Creates an empty IGMP packet with the specified group address
	 * 
	 * @param groupAddress group address for the IGMP report 
	 */
	public IGMPV2LeaveGroupPacket(Inet4Address groupAddress) {
		super();
		this.groupAddress = groupAddress;
	}
	
	/**
	 * Creates a IGMP packet 
	 * 
	 * @param header IPV4 packet header
	 * @param dataBuffer IGMP message
	 */ 
	public IGMPV2LeaveGroupPacket(IPV4PacketHeader header, ByteBuffer dataBuffer) {
		super(header);
		//this.maxResponseTime=dataBuffer.get(); // IGMPv2
		dataBuffer.get(); // field: unused
		dataBuffer.getShort(); // field: checksum
		this.groupAddress=IPV4AddressUtil.readAddress(dataBuffer);
	}
	
	/**
	 * Returns data length of IGMP message
	 * 
	 * @see ch.unibe.iam.mcast.net.ip.IPV4Packet#getDataLengt()
	 * @return data length of IGMP message
	 */
	protected int getDataLengt() {
		return 8; // TODO change me in the IGMPV3 implementation
	}
	
	/**
	 * Gets the group address specified in the IGMP report message
	 
	 * @return Group address specified in the IGMP report message
	 */
	public Inet4Address getGroupAddress() {
		return this.groupAddress;
	}
	
	/**
	 * Serializes the IGMP report specific message data
	 * 
	 * @param buffer Buffer to fill with serialized IGMP message data
	 * @see ch.unibe.iam.mcast.net.ip.IPV4Packet#serializeData(java.nio.ByteBuffer)
	 */
	protected void serializeDataInternal(ByteBuffer buffer) {
		int startPosition=buffer.position();
		buffer.put((byte) 0x12);
		// buffer.put(maxResponseTime); // IGMPv2
		buffer.put((byte)0x00); // set zero for IGMPv1 (unused)
		buffer.putShort((short)0); // checksum
		buffer.put(groupAddress.getAddress()); 
		log.debug("IGMP data successfully serialized for: " + groupAddress.getHostAddress());
	}
	
}
