package ch.unibe.iam.mcast.router;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.omg.CORBA.CTX_RESTRICT_SCOPE;

import ch.unibe.iam.mcast.core.ConnectCallback;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.core.MulticastRoutingTable;
import ch.unibe.iam.mcast.core.PeerConnection;
import ch.unibe.iam.mcast.core.java.IPV4AddressCache;
import ch.unibe.iam.mcast.core.java.JavaMulticastCore;
import ch.unibe.iam.mcast.core.java.JavaTapNetworkInterface;
import ch.unibe.iam.mcast.core.java.RawIPPacketListener;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.router.p2p.ConnectionInformation;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilter;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext;
import ch.unibe.iam.mcast.router.p2p.PeerInstance;
import ch.unibe.iam.mcast.router.p2p.filter.KeepAliveFilter;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;
import ch.unibe.iam.mcast.router.p2p.message.euqos.ChildHelloMessage;
import ch.unibe.iam.mcast.router.p2p.message.euqos.ParentHelloMessage;
import ch.unibe.iam.mcast.router.pastry.GroupPeerDHT;
import ch.unibe.iam.mcast.router.qos.QoSInformation;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;
import ch.unibe.iam.mcast.router.qos.QoSReservationService;
import ch.unibe.iam.mcast.router.qos.QosReservationException;
import ch.unibe.iam.mcast.router.qos.ReservedResources;

public class EUQoSMulticastPeer extends PeerInstance {

	private JavaTapNetworkInterface tapDevice;

	// private final Map<InetAddress,Set<E>> abc;
	// private final Map<InetSocketAddress, InetAddress> rootMap = new
	// HashMap<InetSocketAddress, InetAddress>();
	private final Map<InetAddress, Set<ConnectionInformation>> connectionsByGroup = new HashMap<InetAddress, Set<ConnectionInformation>>();

	private final Map<InetSocketAddress, Set<InetAddress>> groupsToJoin = new HashMap<InetSocketAddress, Set<InetAddress>>();

	private final Map<ConnectionInformation, ReservedResources> reservedResources = new HashMap<ConnectionInformation, ReservedResources>();

	private final String peerId;

	private IGMPProcess igmpProcess;

	private final GroupPeerDHT dht;

	private final RoutingFilter routingFilter;

	private final JoinState joinState = new JoinState();

//	private final QoSInformation info;

	private final QoSReservationService reservationService;

	public EUQoSMulticastPeer(InetSocketAddress bindAddress, InetSocketAddress overlayBindAddress, InetSocketAddress[] bootstrapNodes, String peerId, QoSInformation info,
			QoSReservationService reservationService) throws IOException {
		super(new JavaMulticastCore(bindAddress));
		
		this.reservationService = reservationService;
		routingFilter = new RoutingFilter();
		super.addMessageFilter(new KeepAliveFilter());
		super.addMessageFilter(routingFilter);
		this.peerId = peerId;
		this.dht = new GroupPeerDHT(overlayBindAddress, bootstrapNodes, bindAddress, routingFilter);
	}

	private InetAddress virtualInterfaceAddress() {
		Random r = new Random();
		byte adr[] = { 10, (byte) r.nextInt(256), (byte) r.nextInt(256), (byte) r.nextInt(256) };
		InetAddress ret;
		try {
			ret = InetAddress.getByAddress("random-tap-address", adr);
		} catch (UnknownHostException e) {
			throw new RuntimeException("this should never happend");
		}

		return ret;
	}

	public void start() throws IOException {
		super.start();
		((JavaMulticastCore) super.getCore()).start();
		this.tapDevice = (JavaTapNetworkInterface) getCore().openVirtualInterface(virtualInterfaceAddress());
		this.tapDevice.addRawIPPacketListener(new SenderGroupJoiner());
		igmpProcess = new IGMPProcess(tapDevice);
		igmpProcess.addMulticastMembershipChangeEventListener(routingFilter);
		igmpProcess.start();

		try {
			dht.start();
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		joinState.start();
	}

	public void stop() {
		joinState.stop();
		dht.stop();
		((JavaMulticastCore) super.getCore()).stop();

	}

	private void joinOverlayGroup(final InetAddress addedGroup, final QoSRequirement requirement) {
		System.out.println(">>>>>>>>> joining group " + addedGroup);
		new Thread(new Runnable() {
			public void run() {
				try {
					dht.joinTo(addedGroup, requirement);
				} catch (InterruptedException e) {
				}
				getCore().getRoutingTable().addRoute(addedGroup, tapDevice, MulticastRoutingTable.PRIORITY_MEDIUM);
				System.out.println("routing table: " + getCore().getRoutingTable());

			}
		}).start();
	}

	private void leaveOverlayGroup(final InetAddress group) {
		System.out.println(">>>>>>>>> leaving group " + group);
		Set<ConnectionInformation> conns;
		synchronized (connectionsByGroup) {
			conns = connectionsByGroup.remove(group);
		}
		if (conns != null) {
			for (ConnectionInformation ci : conns) {
				PeerConnection pc = getConnectionFor(ci);
				if (pc != null) {
					pc.close();
				}
			}
		}
		new Thread(new Runnable() {
			public void run() {
				dht.leaveGroup(group);
				getCore().getRoutingTable().removeRoute(group, tapDevice);
				System.out.println("routing table: " + getCore().getRoutingTable());
			}
		}).start();
	}

	private class RoutingFilter implements P2PMessageFilter, OverlayTopologyChangeListener, MulticastMembershipChangeEventListener {

		private P2PMessageFilterContext ctx;

		public boolean handleMessage(ConnectionInformation peerAddress, P2PMessage message) {
			if (message instanceof ParentHelloMessage) {
				ParentHelloMessage hello = (ParentHelloMessage) message;
				InetAddress groupAddress = hello.getGroupAddress();
				System.out.println(">>> parrent hello trough " + peerAddress + " group " + groupAddress);
				ChildHelloMessage response = new ChildHelloMessage(groupAddress, peerId);
				ctx.sendMessage(peerAddress, response);
				getCore().getRoutingTable().addRoute(groupAddress, ctx.getPeerConnectionFor(peerAddress), MulticastRoutingTable.PRIORITY_MEDIUM);
				System.out.println("routing table: " + getCore().getRoutingTable());
				synchronized (connectionsByGroup) {
					Set<ConnectionInformation> conns = connectionsByGroup.get(groupAddress);
					if (conns == null) {
						conns = new HashSet<ConnectionInformation>();
						connectionsByGroup.put(groupAddress, conns);
					}
					conns.add(peerAddress);
				}
				String remoteId = hello.getPeerId();
				reserveResources(peerAddress, groupAddress, remoteId);
				return true;
			} else if (message instanceof ChildHelloMessage) {
				ChildHelloMessage hello = (ChildHelloMessage) message;
				InetAddress groupAddress = hello.getGroupAddress();
				System.out.println(">>> child hello trough " + peerAddress + " group " + groupAddress);
				getCore().getRoutingTable().addRoute(groupAddress, ctx.getPeerConnectionFor(peerAddress), MulticastRoutingTable.PRIORITY_MEDIUM);
				System.out.println("routing table: " + getCore().getRoutingTable());
				String remotePeerId = hello.getPeerId();
				reserveResources(peerAddress, groupAddress, remotePeerId);
				return true;
			} else {
				return false;
			}

		}

		private void reserveResources(final ConnectionInformation peerAddress, InetAddress group, final String remoteId) {
			final QoSRequirement req = joinState.getQosRequirement(group);
			System.out.println("reservation requirements: " + req);
			if (req != null) {

				Thread t = new Thread(new Runnable() {

					public void run() {
						try {
							ReservedResources resources = reservationService.reserveResources(peerAddress.getLocalAddress(), peerAddress.getRemoteAddress(), req);
							synchronized (reservedResources) {
								reservedResources.put(peerAddress, resources);
							}
						} catch (QosReservationException e) {
							log.warn("could not reserve resources for " + peerAddress, e);
						}

					}
				});
				t.start();
			}
		}

		public void onNewPeerConnection(ConnectionInformation peerAddress) {

			System.out.println(">>>>> new peer connection from " + peerAddress);

			InetAddress group;
			// synchronized (EUQoSMulticastPeer.this.rootMap) {
			// group = rootMap.get(peerAddress);
			// }

			synchronized (groupsToJoin) {
				Set<InetAddress> groups = groupsToJoin.get(peerAddress.getRemoteAddress());
				if (groups == null) {
					return;
				}
				Iterator<InetAddress> it = groups.iterator();
				if (it.hasNext()) {
					group = it.next();
					it.remove();
					if (!it.hasNext()) {
						groupsToJoin.remove(peerAddress.getRemoteAddress());
					}
				} else {
					log.warn("iterator was empty");
					return;
				}
			}

			if (group != null) {

				synchronized (connectionsByGroup) {
					Set<ConnectionInformation> conns = connectionsByGroup.get(group);
					if (conns == null) {
						conns = new HashSet<ConnectionInformation>();
						connectionsByGroup.put(group, conns);
					}
					conns.add(peerAddress);
				}

				ParentHelloMessage msg = new ParentHelloMessage(peerId, group);
				ctx.sendMessage(peerAddress, msg);
			}
		}

		public void onPeerConnectionClosed(ConnectionInformation peerAddress) {
			System.out.println(getCore().getRoutingTable());
			synchronized (reservedResources) {
				ReservedResources resources = reservedResources.remove(peerAddress);
				if (resources != null) {
					resources.releaseResources();
				}
			}
			// synchronized(connectionsByGroup) {
			// PeerConnection conn = ctx.getPeerConnectionFor(peerAddress);
			// for (Map.Entry<InetAddress, Set<PeerConnection>> entry :
			// connectionsByGroup.entrySet()) {
			//					
			// }
			// }
			//
			getCore().getRoutingTable().removeDestination(ctx.getPeerConnectionFor(peerAddress));
			System.out.println(getCore().getRoutingTable());
		}

		public void setContext(P2PMessageFilterContext context) {
			this.ctx = context;
		}

		public void start() {
			// TODO Auto-generated method stub

		}

		public void stop() {
			// TODO Auto-generated method stub

		}

		public void onMembershipStateChanged(MulticastMembershipChangeEnvent evt) {
			evt.getAddedGroups();

			for (InetAddress addedGroupPair : evt.getAddedGroups()) {
				joinState.joinIGMP(addedGroupPair);
			}

			for (InetAddress removedGroup : evt.getRemovedGroups()) {
				joinState.leaveIGMP(removedGroup);
			}


		}

		public void addConnection(InetAddress group, InetSocketAddress connection) {
			System.out.println(">>>>>>>> connect to " + connection + " for group " + group);
			synchronized (groupsToJoin) {
				Set<InetAddress> groups = groupsToJoin.get(connection);
				if (groups == null) {
					groups = new HashSet<InetAddress>();
					groupsToJoin.put(connection, groups);
				}
				groups.add(group);
			}
			connectTo(connection, new ConnectCallback() {

				public void connectFailed(InetSocketAddress remoteAddress, IOException reason) {
					// TODO Auto-generated method stub
					
				}

				public void connectionOpen(PeerConnection conn) {
					// TODO Auto-generated method stub
					
				}
				
			});
		}

		public void removeConnection(InetAddress group, InetSocketAddress connection) {

		}

	}

	private class SenderGroupJoiner implements RawIPPacketListener {

		private IPV4AddressCache cache = new IPV4AddressCache();

		public void onNewRawIPPacket(LocalNetworkInterface iface, ByteBuffer frameBuffer) {
			int destination = frameBuffer.getInt(frameBuffer.position() + IPProtocolConstants.IPV4_DESTINATION_ADDRESS_OFFSET);
			final Inet4Address group = cache.lookup(destination);
			if (group.isMulticastAddress()) {
				joinState.joinSender(group);
			}
			// if (joinState.joinSender(group)) {
			// joinOverlayGroup(group);
			// }
			// getCore().getRoutingTable().
		}

	}

	private class JoinState {

		private QoSInformation reservedResources = new QoSInformation();

		private final Set<InetAddress> groupsJoinedByIGMP = new HashSet<InetAddress>();

		private final Map<InetAddress, PacketCounter> groupsJoinedBySender = new HashMap<InetAddress, PacketCounter>();

		private final CleanupJob cleanupJob = new CleanupJob();

		private Thread cleanupThread;

		public synchronized void start() {
			this.cleanupThread = new Thread(cleanupJob);
			this.cleanupThread.start();
		}

		public QoSRequirement getQosRequirement(InetAddress group) {
			return reservedResources.getRequirement(group);
		}

		public synchronized void stop() {
			if (cleanupThread != null)
				this.cleanupThread.interrupt();
		}

		public synchronized void joinIGMP(InetAddress group) {
			boolean joined = isJoined(group);
			groupsJoinedByIGMP.add(group);
			boolean ret = !joined;

			if (ret)
				joinOverlayGroup(group, reservedResources.getRequirement(group));
			// System.out.println("IGMP join (" + group + ")=" + ret);
			// return ret;
		}

		public synchronized void leaveIGMP(InetAddress group) {
			groupsJoinedByIGMP.remove(group);
			boolean joined = groupsJoinedBySender.containsKey(group);

			boolean ret = !joined;
			if (ret) {
				leaveOverlayGroup(group);
			}
			// System.out.println("IGMP leave (" + group + ")=" + ret+"
			// igmpGroups="+groupsJoinedByIGMP+" |
			// senderGroups="+groupsJoinedBySender);
			// return ret;
		}

		public synchronized void joinSender(InetAddress group) {
			boolean joined = isJoined(group);
			PacketCounter counter = groupsJoinedBySender.get(group);
			if (counter == null) {
				counter = new PacketCounter();
				groupsJoinedBySender.put(group, counter);
			}
			counter.sentPacket();
			boolean ret = !joined;
			if (ret) {
				joinOverlayGroup(group,reservedResources.getRequirement(group));
			}
		}

		private synchronized void leaveSender(InetAddress group) {
			groupsJoinedBySender.remove(group);
			boolean joined = groupsJoinedByIGMP.contains(group);
			boolean ret = !joined;
			if (ret) {
				leaveOverlayGroup(group);
			}
		}

		private synchronized boolean isJoined(InetAddress group) {
			return groupsJoinedByIGMP.contains(group) || groupsJoinedBySender.containsKey(group);
		}

		private class PacketCounter {
			private boolean packetsSent;

			private void sentPacket() {
				this.packetsSent = true;
			}

			public boolean arePacketsSent() {
				boolean ret = packetsSent;
				packetsSent = false;
				return ret;
			}

		}

		private class CleanupJob implements Runnable {

			public void run() {
				while (!Thread.interrupted()) {
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e) {
						return;
					}
					synchronized (JoinState.this) {
						Set<InetAddress> toLeave = new HashSet<InetAddress>();
						for (InetAddress group : groupsJoinedBySender.keySet()) {
							PacketCounter counter = groupsJoinedBySender.get(group);
							if (!counter.arePacketsSent()) {
								toLeave.add(group);

							}
						}
						for (InetAddress group : toLeave) {
							leaveSender(group);
						}
					}
				}

			}

		}

		public synchronized void changeQosReservation(InetAddress group, QoSRequirement req) {
			System.out.println("changing "+group.getHostAddress()+" -> "+req);
			QoSRequirement existing=reservedResources.getRequirement(group);
			if (!existing.equals(req)) {
				
				reservedResources.addRequirement(group, req);				
				if (isJoined(group)) {
					leaveOverlayGroup(group);
					joinOverlayGroup(group, req);
				}
			}
			
		}
	}

	public MulticastRoutingTable getMulticastRoutingTable() {
		return super.getCore().getRoutingTable();
	}

	public void changeQoSReservation(InetAddress group, QoSRequirement req) {
		
		joinState.changeQosReservation(group,req);
		
//		QoSRequirement currentReq = info.getRequirement(group);
//		if (!req.equals(currentReq)) {
//			info.addRequirement(group, req);
//			leaveOverlayGroup(group);
//			joinOverlayGroup(group);
//		}
	}

}
