/*
 * Created on Oct 26, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package ch.unibe.iam.mcast.router;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.EventListener;
import java.util.EventObject;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.IPV4PacketListener;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPPacket;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPV1QueryPacket;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPV1ReportPacket;
import ch.unibe.iam.mcast.util.EventListenerCollection;
import ch.unibe.iam.mcast.util.FireEventMethod;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window - Preferences - Java - Code Style - Code Templates
 */
public class IGMPProcess implements IPV4PacketListener, Runnable {
    
    
    
    
    public static int DEFAULT_TTL = 2;
    private static final Inet4Address IGMP_ROUTER_ADDRESS;
    static {
        try {
            IGMP_ROUTER_ADDRESS = (Inet4Address) Inet4Address.getByAddress(new byte[] { 10, 1, 1, 1 });
        } catch (UnknownHostException e) {
            throw new RuntimeException(e);
        }

    }

    public static int DEFAULT_DELAY = 15000;

    private final LocalNetworkInterface iface;

    // private final InterfaceInformation info;

    private Thread myThread;

    /**
     * Max time to life for entries in the group table
     *  
     */
    protected int groupTTL = DEFAULT_TTL;

    /**
     * Stores the group table entries key: multi cast group, data: TTL
     *  
     */
    protected HashMap<InetAddress,Integer> groupTable;
    
   

    /**
     * Stores the event listener list
     */
    private final EventListenerCollection membershipListeners = new EventListenerCollection();

    private static final Log log = LogFactory.getLog(IGMPProcess.class);

    /**
     * 
     */
    public IGMPProcess(LocalNetworkInterface iface) {
        groupTable = new HashMap<InetAddress,Integer>();
        this.iface = iface;
        iface.addIPV4PacketListener(IPProtocolConstants.IGMP, this);
        
        boolean shouldLog = RouterUtil.performLogging();
		if (shouldLog) {
        	addMulticastMembershipChangeEventListener(new IGMPEventLogger());
        }
        
    }

	/*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.iface.PacketReceiver#onPacket(ch.unibe.iam.mcast.ip.IPV4Packet)
     */
    public void onIPPacket(LocalNetworkInterface iface,IPV4Packet packet) {
        if (packet instanceof IGMPPacket) {
            if (packet instanceof IGMPV1QueryPacket) {
                // TODO handle the query etc.
            } else if (packet instanceof IGMPV1ReportPacket) {
                IGMPV1ReportPacket report = (IGMPV1ReportPacket) packet;
                Inet4Address groupAddress = report.getGroupAddress();
                put(groupAddress);
            }
        } else {
            if (log.isWarnEnabled()) {
                log.warn("received non-igmp packet: " + packet);
            }
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Runnable#run()
     */
    public void run() {
        while (!Thread.interrupted()) {
            try {
                IGMPV1QueryPacket query = new IGMPV1QueryPacket();
                query.setSourceAddress(IGMP_ROUTER_ADDRESS);
                iface.sendIPV4Packet(query,0);
                updateTable();
                Thread.sleep(DEFAULT_DELAY);
            } catch (InterruptedException e) {
                // just ignore and exit
                return;
            }
        }

    }

    public synchronized void start() {
        this.myThread = new Thread(this);
        myThread.start();
    }

    public void stop() {
        myThread.interrupt();
    }
    
    /**
     * @param newGroups
     * @param removedGroups
     */
    protected void fireMulticastMembershipChangeEvent(Set<InetAddress> newGroups, Set<InetAddress> removedGroups) {
        MulticastMembershipChangeEnvent event = new MulticastMembershipChangeEnvent(this, iface, newGroups, removedGroups);
        membershipListeners.fireEvent(new FireEventMethod() {

            public void fireEvent(EventListener listener, EventObject object) {
                MulticastMembershipChangeEventListener l = (MulticastMembershipChangeEventListener) listener;
                l.onMembershipStateChanged((MulticastMembershipChangeEnvent) object);
            }
        }, event);
    }

    /**
     * Puts a group address to the group table and renews its TTL if already
     * stored
     * 
     * @param inet4address
     *            Group address to store / renew
     */
    public synchronized void put(InetAddress inet4address) {
        if (!groupTable.containsKey(inet4address)) {
            Set<InetAddress> newGroups = Collections.singleton(inet4address);
            Set<InetAddress> removedGroups = Collections.emptySet();
            
            fireMulticastMembershipChangeEvent(newGroups, removedGroups);
        }
        groupTable.put(inet4address, new Integer(groupTTL));
        if (log.isDebugEnabled())
            log.debug("new multicast group address added: " + inet4address.getHostAddress() + " with TTL " + groupTTL);
    }

    /**
     * Checks wether a group address is already stored in the group table
     * 
     * @param inet4address
     *            Group address to check
     * @return <code>true</code> if already in group table, else
     *         <code>false</code>
     */
    public boolean check(Inet4Address inet4address) {
        if (log.isDebugEnabled())
            log.debug("checking if " + inet4address.getHostAddress() + " is in multicast group table");
        if (groupTable.containsKey(inet4address))
            return true;
        else
            return false;
    }

    /**
     * Iterates through all group table entries, decreses TTL and removes
     * entries with TTL=0
     *  
     */
    public void updateTable() {
        if (log.isDebugEnabled()) {
            log.debug("entries in multicast table before update: " + groupTable.size());
        }
        
        HashSet<InetAddress> removed = new HashSet<InetAddress>();
        for (Iterator<Map.Entry<InetAddress, Integer>> i = groupTable.entrySet().iterator(); i.hasNext();) {
        	Map.Entry<InetAddress, Integer> entry=i.next();
            InetAddress inetAddress = entry.getKey();
            Integer ttl = entry.getValue();
            //System.out.println("ttl: "+ttl);
            if (ttl.intValue() < 1) {
                i.remove();
                removed.add(inetAddress);
            } else {
                entry.setValue(new Integer(ttl.intValue() - 1));
            }                
        }
    
        if (log.isDebugEnabled()) {
            log.debug("entries in multicast table after update: " + groupTable.size());
        }
        
        Set<InetAddress> empty=Collections.emptySet();
        
        if (!removed.isEmpty()) { 
        	
            fireMulticastMembershipChangeEvent(empty, removed);
        }
    }

    /**
     * Checks if the group table is empty
     * 
     * @return <code>true</code> if empty, else <code>false</code>
     */
    public boolean isEmpty() {
        return groupTable.isEmpty();
    }

    /**
     * Adds an membership event listener.
     * 
     * @param listener
     *            The listener to be added.
     */
    public void addMulticastMembershipChangeEventListener(MulticastMembershipChangeEventListener listener) {
        membershipListeners.addEventListener(listener);
    }

    /**
     * Removes an membership event listener.
     * 
     * @param listener
     *            The listener to be removed.
     */
    public void removeMulticastMembershipChangeEventListener(MulticastMembershipChangeEventListener listener) {
        membershipListeners.removeEventListener(listener);
    }
    
    
}