package ch.unibe.iam.mcast.router;

import java.net.Inet4Address;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import ch.unibe.iam.mcast.core.IPV4PacketListener;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.core.MulticastRoutingTable;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPPacket;
import ch.unibe.iam.mcast.net.ip.igmp.IGMPV2QueryPacket;

public class IGMPV2GroupManager implements IPV4PacketListener {
	
	private enum STATE {NO_MEMBERS,MEMBER_PRESENT,CHECKING_MEMBERSHIP};
	private final Map<Inet4Address, GroupState> groupStates=new HashMap<Inet4Address, GroupState>();
	
	private final Timer timer=new Timer("IGMP process Timer",true);

	private final LocalNetworkInterface iface;

	private final float QUERY_TIMEOUT=10.0f;
	
	public IGMPV2GroupManager(LocalNetworkInterface iface) {
		this.iface = iface;
		iface.addIPV4PacketListener(IPProtocolConstants.IGMP, this);
	}
	
	public void stop() {
		timer.cancel();
		iface.removeIPV4PacketListener(IPProtocolConstants.IGMP, this);
	}

	
	private void handlePacket(IGMPPacket packet) {
		// TODO Auto-generated method stub
		
	}

	
	private class GroupState {
		private final Inet4Address group;
		private STATE state;
		
		
		public GroupState(Inet4Address group) {
			this.group = group;
			this.state=STATE.NO_MEMBERS;
		}
		
		
		public void reportReceived() {
			if (state == STATE.NO_MEMBERS) {
				
			} else if (state == STATE.MEMBER_PRESENT) {
				
			} else {
				
			}
		}
		
		private class QueryGroupTask extends TimerTask {

			@Override
			public void run() {
				
				
			}
			
		}
	}
	

	private void sendQuery(Inet4Address groupAddress) {
		iface.sendIPV4Packet(groupAddress == null ? new IGMPV2QueryPacket(QUERY_TIMEOUT) : new IGMPV2QueryPacket(QUERY_TIMEOUT,groupAddress), MulticastRoutingTable.PRIORITY_HIGH);
	}

	public void onIPPacket(LocalNetworkInterface source, IPV4Packet packet) {
		if (packet instanceof IGMPPacket) {
			handlePacket((IGMPPacket)packet);
		}
		
	}


}
