/*
 * Created on Nov 11 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package ch.unibe.iam.mcast.router;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.ConnectCallback;
import ch.unibe.iam.mcast.core.LocalNetworkInterface;
import ch.unibe.iam.mcast.core.MulticastRoutingDestination;
import ch.unibe.iam.mcast.core.MulticastRoutingTable;
import ch.unibe.iam.mcast.core.MulticastRoutingTableEntry;
import ch.unibe.iam.mcast.core.PeerConnection;
import ch.unibe.iam.mcast.core.java.IPV4AddressCache;
import ch.unibe.iam.mcast.core.java.JavaMulticastCore;
import ch.unibe.iam.mcast.core.java.JavaTapNetworkInterface;
import ch.unibe.iam.mcast.core.java.RawIPPacketListener;
import ch.unibe.iam.mcast.net.ip.IPProtocolConstants;
import ch.unibe.iam.mcast.router.p2p.ConnectionInformation;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilter;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext;
import ch.unibe.iam.mcast.router.p2p.PeerInstance;
import ch.unibe.iam.mcast.router.p2p.filter.KeepAliveFilter;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.AddRouteMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.OpenMessage;
import ch.unibe.iam.mcast.router.p2p.message.mcast.RemoveRouteMessage;
import ch.unibe.iam.mcast.router.qos.QoSInformation;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;
import ch.unibe.iam.mcast.router.qos.QoSReservationService;
import ch.unibe.iam.mcast.router.qos.QosReservationException;
import ch.unibe.iam.mcast.router.qos.ReservedResources;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window -
 * Preferences - Java - Code Style - Code Templates
 */
public class QoSMulticastPeer implements MulticastMembershipChangeEventListener {

	private static final long RECONNECT_TIMEOUT = 1000;

	private static final Log log = LogFactory.getLog(QoSMulticastPeer.class);

	private final PeerInstance peerInstance;

	private final InetSocketAddress masterNode;

	private final MulticastRoutingTable routingTable;

	private final MulticastFilter multicastFilter;

	private final JavaMulticastCore core;

	private LocalNetworkInterface localInterface;

//	private final InetSocketAddress peerSocketAddress;

	private final QoSInformation qosInfo;

	private final QoSReservationService reservationService;

	private final Timer timer = new Timer("QoSMulticastPeer Timer", true);

	private boolean isMaster;

	private JoinState joinState = new JoinState();

	// private Set<InetAddress> activeConnects=new HashSet<InetAddress>();

	Map<ConnectionInformation, GroupAndResources> overlayConnections = new HashMap<ConnectionInformation, GroupAndResources>();

	/**
	 * @throws IOException
	 * 
	 */
	public QoSMulticastPeer(InetSocketAddress localAddress, InetSocketAddress masterNode, Inet4Address interfaceAddress, QoSInformation qosInfo,
			QoSReservationService reservationService) throws IOException {
//		this.peerSocketAddress = localAddress;
		this.masterNode = masterNode;
		this.qosInfo = qosInfo;
		this.reservationService = reservationService;

		this.core = new JavaMulticastCore(localAddress);
		this.routingTable = core.getRoutingTable();

		this.peerInstance = new PeerInstance(core);
		this.peerInstance.addMessageFilter(new KeepAliveFilter());
		this.multicastFilter = new MulticastFilter();
		// routingTable.addMulticastRoutingTableListener(multicastFilter);
		this.peerInstance.addMessageFilter(multicastFilter);

		peerInstance.addMessageFilter(new RandomAddressFilter());

	}

	/**
	 * @param interfaceAddress
	 * @throws IOException
	 */
	private void openVirtualInterface(Inet4Address interfaceAddress) throws IOException {
		if (this.localInterface != null) {
			log.error("trying to initialize already initialized local interface");
		}
		LocalNetworkInterface iface = core.openVirtualInterface(interfaceAddress);
		IGMPProcess igmpProcess = new IGMPProcess(iface);
		igmpProcess.addMulticastMembershipChangeEventListener(this);
		igmpProcess.start();

		((JavaTapNetworkInterface) iface).addRawIPPacketListener(new SenderGroupJoiner());

		this.localInterface = iface;

	}

	public void start() throws IOException {
		core.start();
		peerInstance.start();
		joinState.start();
		this.isMaster = isMasterNode();

	}

	/**
	 * @return
	 * @throws SocketException
	 */
	private boolean isMasterNode() throws SocketException {
		boolean isMaster = false;
		for (Enumeration interfaces = NetworkInterface.getNetworkInterfaces(); interfaces.hasMoreElements();) {
			NetworkInterface ni = (NetworkInterface) interfaces.nextElement();
			for (Enumeration e = ni.getInetAddresses(); e.hasMoreElements();) {
				InetAddress a = (InetAddress) e.nextElement();
				isMaster |= a.equals(masterNode.getAddress());
			}

		}
		return isMaster;
	}

	public void stop() throws IOException {
		timer.cancel();
		peerInstance.stop();
		core.stop();
		joinState.stop();
		// if (!isMasterNode()) {
		// // this.connectToMaster.interrupt();
		// // try {
		// // this.connectToMaster.join();
		// // } catch (InterruptedException e) {
		// // log.error(e);
		// // }
		// }

	}

	// public void connectTo(InetSocketAddress peerAddress) {
	// peerInstance.connectTo(peerAddress);
	// }

	public void onMembershipStateChanged(MulticastMembershipChangeEnvent evt) {

		for (Iterator i = evt.getAddedGroups().iterator(); i.hasNext();) {
			final InetAddress group = (InetAddress) i.next();

			joinState.joinIGMP(group);
		}

		for (Iterator i = evt.getRemovedGroups().iterator(); i.hasNext();) {

			InetAddress group = (InetAddress) i.next();
			joinState.leaveIGMP(group);

		}

	}

	private void leaveOverlayGroup(InetAddress group) {
		// routingTable.removeRoute(group, localInterface);
		if (!isMaster) {
			Set<MulticastRoutingTableEntry> destinations = routingTable.getRoutes().get(group);
			for (MulticastRoutingTableEntry ent : destinations) {
				MulticastRoutingDestination dst = ent.getDestination();
				if (dst instanceof PeerConnection) {
					PeerConnection pc = (PeerConnection) dst;
					// pc.sendMessage(new RemoveRouteMessage(group));
					pc.close();
				}
			}
		}
	}

	private boolean stillSubscribed(final InetAddress group) {
		return (routingTable.getRoutes().get(group) != null) && (!routingTable.getRoutes().get(group).isEmpty());
	}

	private boolean masterConnected(final InetAddress group) {
		Set<MulticastRoutingTableEntry> entries = routingTable.getRoutes().get(group);
		for (MulticastRoutingTableEntry ent : entries) {
			if (ent.getDestination() instanceof PeerConnection) {
				return true;
			}
		}
		return false;
	}

	private void joinOverlayGroup(final InetAddress group, QoSRequirement requirement) {

		routingTable.addRoute(group, localInterface, MulticastRoutingTable.PRIORITY_MEDIUM);

		connectToMaster(group);
	}

	private void connectToMaster(final InetAddress group) {
		if (!isMaster && !masterConnected(group)) {

			RouterUtil.performanceLog("OVERLAY", "connection to master for " + group.getHostAddress());

			System.out.println("connecting to " + masterNode + "...");

			peerInstance.connectTo(masterNode, new ConnectCallback() {

				public void connectFailed(InetSocketAddress remoteAddress, IOException reason) {
					System.out.println("connecting to " + masterNode + " failed, reason: " + reason);
					final ConnectCallback t = this;
					timer.schedule(new TimerTask() {
						@Override
						public void run() {
							if (stillSubscribed(group)) {
								System.out.println("reconnecting to "+masterNode+" for "+group.getHostAddress());
								peerInstance.connectTo(masterNode, t);
							} else {
								System.out.println("cancelling connecting to master for " + group);
							}
						}

					}, RECONNECT_TIMEOUT);

				}

				public void connectionOpen(PeerConnection conn) {

					RouterUtil.performanceLog("OVERLAY", "connected to master for " + group.getHostAddress());

					if (stillSubscribed(group) && !masterConnected(group)) {
						conn.sendMessage(new OpenMessage((short) 0));
						conn.sendMessage(new AddRouteMessage(group, MulticastRoutingTable.PRIORITY_MEDIUM));
						routingTable.addRoute(group, conn, MulticastRoutingTable.PRIORITY_MEDIUM);
						System.out.println("connected to " + masterNode);
						System.out.println("routing table: " + routingTable);
						QoSRequirement req = qosInfo.getRequirement(group);
						ConnectionInformation ci = conn.getConnectionInformation();
						ReservedResources forw = null;
						ReservedResources backw = null;
						try {
							forw = reservationService.reserveResources(ci.getLocalAddress(), ci.getRemoteAddress(), req);
							backw = reservationService.reserveResources(ci.getRemoteAddress(), ci.getLocalAddress(), req);
						} catch (QosReservationException e) {
							log.warn(e);
						}
						synchronized (overlayConnections) {
							overlayConnections.put(conn.getConnectionInformation(), new GroupAndResources(group, forw, backw));
						}
					} else {
						System.out.println("not subscribed anymore, disconnecting");
						conn.close();
					}
				}

			});
		}
	}

	private static class GroupAndResources {
		private final InetAddress group;

		private ReservedResources forwards;

		private ReservedResources backwards;

		public GroupAndResources(InetAddress group, ReservedResources forwards, ReservedResources backwards) {
			this.group = group;
			this.forwards = forwards;
			this.backwards = backwards;

		}

		public ReservedResources getBackwards() {
			return backwards;
		}

		public ReservedResources getForwards() {
			return forwards;
		}

		public InetAddress getGroup() {
			return group;
		}

		public void setForwards(ReservedResources resources) {
			this.forwards = resources;

		}

		public void setBackwards(ReservedResources resources) {
			this.backwards = resources;

		}

	}

	public class MulticastFilter implements P2PMessageFilter {

		private P2PMessageFilterContext context;

		// private Set<ConnectionInformation> multicastPeers = new
		// HashSet<ConnectionInformation>();

		public boolean handleMessage(ConnectionInformation peerAddress, P2PMessage message) {
			System.out.println("received message: " + message);
			if (message instanceof OpenMessage) {
				OpenMessage open = (OpenMessage) message;

				// addMulticastPeer(peerAddress);
				return true;
			} else if (message instanceof AddRouteMessage) {
				AddRouteMessage addRoute = (AddRouteMessage) message;

				InetAddress group = addRoute.getGroupAddress();
				PeerConnection pc = peerInstance.getConnectionFor(peerAddress);

				synchronized (overlayConnections) {
					overlayConnections.put(peerAddress, new GroupAndResources(group, null, null));
				}

				if (pc != null) {
					routingTable.addRoute(group, pc, addRoute.getPriority());
					System.out.println("new routing table: " + routingTable);
				} else {
					if (log.isWarnEnabled())
						log.warn("could not find remote peer data address for the peer " + peerAddress);
				}
				return true;

			} else if (message instanceof RemoveRouteMessage) {
				RemoveRouteMessage removeRoute = (RemoveRouteMessage) message;

				InetAddress group = removeRoute.getGroupAddress();
				PeerConnection pc = peerInstance.getConnectionFor(peerAddress);

				if (pc != null) {
					routingTable.removeRoute(group, pc);
					System.out.println("new routing table: " + routingTable);
				} else {
					if (log.isWarnEnabled())
						log.warn("could not find remote peer data address for the peer " + peerAddress);
				}

				return true;
			}
			return false;
		}

		// private void addMulticastPeer(ConnectionInformation peerAddress) {
		// boolean isNewPeer;
		// synchronized (this) {
		// isNewPeer = multicastPeers.add(peerAddress);
		// }
		// if (isNewPeer) {
		// for (Iterator i = routingTable.getRoutes().keySet().iterator();
		// i.hasNext();) {
		// InetAddress group = (Inet4Address) i.next();
		// if (routingTable.hasRoutesForGroup(group)) {
		// int priority = routingTable.getMaxPriority(group);
		// AddRouteMessage msg = new AddRouteMessage(group, priority);
		// context.sendMessage(peerAddress, msg);
		// }
		//
		// }
		// }
		//
		// }

		// private synchronized void removeMulticastPeer(ConnectionInformation
		// peerAddress) {
		// multicastPeers.remove(peerAddress);
		// }

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#onNewPeerConnection(java.net.InetSocketAddress)
		 */
		public void onNewPeerConnection(ConnectionInformation peerAddress) {
			System.out.println("new peer connection: " + peerAddress);
			// context.sendMessage(peerAddress, new OpenMessage((short) 0));
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#onPeerConnectionClosed(java.net.InetSocketAddress)
		 */
		public void onPeerConnectionClosed(ConnectionInformation peerAddress) {
			GroupAndResources groupAndResources;

			synchronized (overlayConnections) {
				groupAndResources = overlayConnections.remove(peerAddress);
			}

			if (groupAndResources != null) {

				if (groupAndResources.getForwards() != null) {
					groupAndResources.getForwards().releaseResources();
				}

				if (groupAndResources.getBackwards() != null) {
					groupAndResources.getBackwards().releaseResources();
				}

				PeerConnection pc = peerInstance.getConnectionFor(peerAddress);
				if (pc != null) {
					routingTable.removeDestination(pc);
				}
				connectToMaster(groupAndResources.getGroup());
			} else {
				log.warn("closed connection " + peerAddress + " for unknown group");
			}

			// removeMulticastPeer(peerAddress);
			//
			// PeerConnection conn = peerInstance.getConnectionFor(peerAddress);
			//
			// if (conn == null) {
			// log.error("no connection to the peer !?");
			// }
			// routingTable.removeDestination(conn);

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#setContext(ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext)
		 */
		public void setContext(P2PMessageFilterContext context) {
			this.context = context;

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#start()
		 */
		public void start() {
			System.out.println("multicast filter ready");
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#stop()
		 */
		public void stop() {

		}

		// public void multicastRouteAdded(MulticastRoutingTable table,
		// InetAddress group, MulticastRoutingTableEntry entry) {
		// System.out.println("added mcast route: " + group + " -> " + entry);
		// AddRouteMessage addRoute = new AddRouteMessage(group,
		// table.getMaxPriority(group));
		//
		// sendToAllPeerExceptSource(addRoute, entry.getDestination());
		//
		// }

		/**
		 * @param addRoute
		 * @param source
		 */
		// private void sendToAllPeerExceptSource(P2PMessage message,
		// MulticastRoutingDestination source) {
		// Set<ConnectionInformation> allPeers = getMulticastPeers();
		// PeerConnection conn = source instanceof PeerConnection ?
		// (PeerConnection) source : null;
		// for (ConnectionInformation peerAddress : allPeers) {
		// if (!(conn != null &&
		// conn.getConnectionInformation().equals(peerAddress))) {
		// context.sendMessage(peerAddress, message);
		// }
		// }
		// }
		//
		// private synchronized Set<ConnectionInformation> getMulticastPeers() {
		// return new HashSet<ConnectionInformation>(multicastPeers);
		// }
		//
		// public void multicastRouteRemoved(MulticastRoutingTable table,
		// InetAddress group, MulticastRoutingTableEntry entry) {
		// System.out.println("removed mcast route: " + group + " -> " + entry);
		// if (!table.hasRoutesForGroup(group)) {
		//
		// RemoveRouteMessage removeRoute = new RemoveRouteMessage(group);
		// sendToAllPeerExceptSource(removeRoute, entry.getDestination());
		//
		// }
		//
		// }
	}

	private class RandomAddressFilter implements P2PMessageFilter {

		private P2PMessageFilterContext ctx;

		private Random random = new Random();

		public boolean handleMessage(ConnectionInformation peerAddress, P2PMessage message) {
			return false;
		}

		public void onNewPeerConnection(ConnectionInformation peerAddress) {
		}

		public void onPeerConnectionClosed(ConnectionInformation peerAddress) {

		}

		public void setContext(P2PMessageFilterContext context) {
			this.ctx = context;
		}

		public void start() {
			byte[] addressArray = new byte[] { 10, (byte) random.nextInt(255), (byte) random.nextInt(255), (byte) random.nextInt(255) };
			Inet4Address address;
			try {
				address = (Inet4Address) InetAddress.getByAddress(addressArray);
			} catch (UnknownHostException e) {
				throw new RuntimeException("this should never happend", e);
			}
			try {
				openVirtualInterface((Inet4Address) address);
			} catch (IOException e) {
				throw new RuntimeException("could not open interface", e);
			}
		}

		public void stop() {
			// TODO Auto-generated method stub

		}

	}

	// private boolean isConnectionToMaster(InetSocketAddress peerAddress) {
	// return peerAddress.equals(masterNode);
	// }

	public MulticastRoutingTable getMulticastRoutingTable() {
		return routingTable;

	}

	public void changeQoSReservation(InetAddress group, QoSRequirement req) {
		QoSRequirement oldReq = qosInfo.getRequirement(group);

		System.out.println("changing QoS for " + group.getHostAddress() + " oldQoS: " + oldReq + " new QoS: " + req);

		// if (isMaster) {
		// return;
		// }

		if (!req.equals(oldReq)) {

			qosInfo.addRequirement(group, req);
			synchronized (overlayConnections) {
				for (Map.Entry<ConnectionInformation, GroupAndResources> ent : overlayConnections.entrySet()) {
					System.out.println("entry: " + ent.getKey() + " " + ent.getValue().getGroup());
					System.out.println("group: " + group);
					if (group.equals(ent.getValue().getGroup())) {
						System.out.println("bingo!");
						ConnectionInformation ci = ent.getKey();
						GroupAndResources gar = ent.getValue();
						changeReservation(group, req, ci, gar);
					}
				}
			}

			
//			if (ci != null) {
//
//			} else {
//				log.error("could not find connection information!");
//			}
		} else {
			log.info("requirements are equal");
		}

	}

	private void changeReservation(InetAddress group, QoSRequirement req, ConnectionInformation ci, GroupAndResources gar) {
		if (ci.getLocalAddress().getPort() != core.getServerPort()) {

			if (gar.getForwards() != null) {
				gar.getForwards().releaseResources();
			}

			if (gar.getBackwards() != null) {
				gar.getBackwards().releaseResources();
			}

			try {
				RouterUtil.performanceLog("RESERVE", "reserving forwards QoS for " + group.getHostAddress());
				gar.setForwards(reservationService.reserveResources(ci.getLocalAddress(), ci.getRemoteAddress(), req));
				RouterUtil.performanceLog("RESERVE", "reserved forwards QoS for " + group.getHostAddress());
			} catch (QosReservationException e) {
				gar.setForwards(null);
				log.error(e);
			}
			try {
				RouterUtil.performanceLog("RESERVE", "reserving backwards QoS for " + group.getHostAddress());
				gar.setBackwards(reservationService.reserveResources(ci.getRemoteAddress(), ci.getLocalAddress(), req));
				RouterUtil.performanceLog("RESERVE", "reserved backwards QoS for " + group.getHostAddress());
			} catch (QosReservationException e) {
				log.error(e);
				gar.setBackwards(null);
			}
		}
	}

	private class SenderGroupJoiner implements RawIPPacketListener {

		private IPV4AddressCache cache = new IPV4AddressCache();

		public void onNewRawIPPacket(LocalNetworkInterface iface, ByteBuffer frameBuffer) {
			int destination = frameBuffer.getInt(frameBuffer.position() + IPProtocolConstants.IPV4_DESTINATION_ADDRESS_OFFSET);
			final Inet4Address group = cache.lookup(destination);
			if (group.isMulticastAddress()) {
				joinState.joinSender(group);
			}
			// if (joinState.joinSender(group)) {
			// joinOverlayGroup(group);
			// }
			// getCore().getRoutingTable().
		}

	}

	private class JoinState {

		private QoSInformation reservedResources = new QoSInformation();

		private final Set<InetAddress> groupsJoinedByIGMP = new HashSet<InetAddress>();

		private final Map<InetAddress, PacketCounter> groupsJoinedBySender = new HashMap<InetAddress, PacketCounter>();

		private final CleanupJob cleanupJob = new CleanupJob();

		private Thread cleanupThread;

		public synchronized void start() {
			this.cleanupThread = new Thread(cleanupJob);
			this.cleanupThread.start();
		}

		public QoSRequirement getQosRequirement(InetAddress group) {
			return reservedResources.getRequirement(group);
		}

		public synchronized void stop() {
			if (cleanupThread != null)
				this.cleanupThread.interrupt();
		}

		public synchronized void joinIGMP(InetAddress group) {
			boolean joined = isJoined(group);
			groupsJoinedByIGMP.add(group);
			boolean ret = !joined;

			if (ret)
				joinOverlayGroup(group, reservedResources.getRequirement(group));
			// System.out.println("IGMP join (" + group + ")=" + ret);
			// return ret;
		}

		public synchronized void leaveIGMP(InetAddress group) {
			groupsJoinedByIGMP.remove(group);
			boolean joined = groupsJoinedBySender.containsKey(group);

			boolean ret = !joined;
			if (ret) {
				leaveOverlayGroup(group);
			}
			// System.out.println("IGMP leave (" + group + ")=" + ret+"
			// igmpGroups="+groupsJoinedByIGMP+" |
			// senderGroups="+groupsJoinedBySender);
			// return ret;
		}

		public synchronized void joinSender(InetAddress group) {
			boolean joined = isJoined(group);
			PacketCounter counter = groupsJoinedBySender.get(group);
			if (counter == null) {
				counter = new PacketCounter();
				groupsJoinedBySender.put(group, counter);
			}
			counter.sentPacket();
			boolean ret = !joined;
			if (ret) {
				joinOverlayGroup(group, reservedResources.getRequirement(group));
			}
		}

		private synchronized void leaveSender(InetAddress group) {
			groupsJoinedBySender.remove(group);
			boolean joined = groupsJoinedByIGMP.contains(group);
			boolean ret = !joined;
			if (ret) {
				leaveOverlayGroup(group);
			}
		}

		private synchronized boolean isJoined(InetAddress group) {
			return groupsJoinedByIGMP.contains(group) || groupsJoinedBySender.containsKey(group);
		}

		private class PacketCounter {
			private boolean packetsSent;

			private void sentPacket() {
				this.packetsSent = true;
			}

			public boolean arePacketsSent() {
				boolean ret = packetsSent;
				packetsSent = false;
				return ret;
			}

		}

		private class CleanupJob implements Runnable {

			private static final int SENDER_JOIN_TIMEOUT = 10000;

			public void run() {
				while (!Thread.interrupted()) {
					try {
						Thread.sleep(SENDER_JOIN_TIMEOUT);
					} catch (InterruptedException e) {
						return;
					}
					synchronized (JoinState.this) {
						Set<InetAddress> toLeave = new HashSet<InetAddress>();
						for (InetAddress group : groupsJoinedBySender.keySet()) {
							PacketCounter counter = groupsJoinedBySender.get(group);
							if (!counter.arePacketsSent()) {
								toLeave.add(group);

							}
						}
						for (InetAddress group : toLeave) {
							leaveSender(group);
						}
					}
				}

			}

		}
	}

	public QoSRequirement getQosReservation(InetAddress group) {
		return qosInfo.getRequirement(group);
		
	}
}