/*
 * Created on Nov 9, 2004
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package ch.unibe.iam.mcast.router.p2p;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.core.ConnectCallback;
import ch.unibe.iam.mcast.core.ConnectionStatusEventListener;
import ch.unibe.iam.mcast.core.MulticastCore;
import ch.unibe.iam.mcast.core.PeerConnection;
import ch.unibe.iam.mcast.core.PeerConnectionListener;
import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window -
 * Preferences - Java - Code Style - Code Templates
 */
public class PeerInstance implements PeerConnectionListener, ConnectionStatusEventListener {

	private final MulticastCore core;

	public static final Log log = LogFactory.getLog(PeerInstance.class);

	private final List<P2PMessageFilter> messageFilters = new ArrayList<P2PMessageFilter>();

	private final Map<ConnectionInformation, PeerConnection> peerConnections = new HashMap<ConnectionInformation, PeerConnection>();

	/**
	 * 
	 */
	public PeerInstance(MulticastCore core) {
		this.core = core;
		core.addConnectionStatusEventListener(this);
	}

	public void connectTo(InetSocketAddress peerAddress, final ConnectCallback callback) {
		core.connectToPeer(peerAddress, callback);

	}

	public void onNewPeerConnectionOpen(PeerConnection connection) {

		
		connection.addPeerMessageListener(this);

		ConnectionInformation connectionInformation = connection.getConnectionInformation();

		synchronized (peerConnections) {
			peerConnections.put(connectionInformation, connection);
		}

		for (int i = 0; i < messageFilters.size(); i++) {
			P2PMessageFilter filter = (P2PMessageFilter) messageFilters.get(i);
			try {
				filter.onNewPeerConnection(connectionInformation);
			} catch (RuntimeException re) {
				log.error("filter thrown runtime exception", re);
			}
		}

	}

	/**
	 * @param msg
	 */
	private void onNewMessage(P2PMessage message, ConnectionInformation connectionInformation) {

		for (int i = 0; i < messageFilters.size(); i++) {
			P2PMessageFilter filter = (P2PMessageFilter) messageFilters.get(i);
			try {
				if (filter.handleMessage(connectionInformation, message)) {
					break;
				}
			} catch (RuntimeException re) {
				log.error("filter thrown runtime exception", re);
			}
		}
	}

	public PeerConnection getConnectionFor(ConnectionInformation connectionInformation) {
		synchronized (peerConnections) {
			return (PeerConnection) peerConnections.get(connectionInformation);
		}
	}

	/**
	 * @param connectionInformation
	 */
	private void connectionClosing(ConnectionInformation connectionInformation) {
		// log.warn("closed connection "+peer);
		PeerConnection removed;
		synchronized (peerConnections) {

			if (!peerConnections.containsKey(connectionInformation)) {
				if (log.isWarnEnabled())
					log.warn("unknown key for closing " + connectionInformation);
			}

			for (int i = 0; i < messageFilters.size(); i++) {
				P2PMessageFilter filter = (P2PMessageFilter) messageFilters.get(i);
				try {
					filter.onPeerConnectionClosed(connectionInformation);
				} catch (RuntimeException re) {
					log.error("filter thrown runtime exception", re);
				}
			}
			removed = (PeerConnection) peerConnections.remove(connectionInformation);

		}
		removed.close();
	}

	public synchronized void start() throws IOException {

		for (Iterator i = messageFilters.iterator(); i.hasNext();) {
			P2PMessageFilter filter = (P2PMessageFilter) i.next();
			try {
				filter.start();
			} catch (RuntimeException e) {
				log.error("runtime exception while starting filter", e);
			}
		}

		System.out.println("peer instance started");

	}

	public synchronized void stop() {

		for (Iterator i = messageFilters.iterator(); i.hasNext();) {
			P2PMessageFilter filter = (P2PMessageFilter) i.next();
			try {
				filter.stop();
			} catch (RuntimeException e) {
				log.error("runtime exception while stopping filter", e);
			}
		}

	}

	public void addMessageFilter(P2PMessageFilter filter) {
		if (filter == null) {
			throw new IllegalArgumentException("message filter should not be null");
		}
		this.messageFilters.add(filter);
		filter.setContext(new MyContext());
	}

	private class MyContext implements P2PMessageFilterContext {

		public void sendMessage(ConnectionInformation connectionInformation, P2PMessage message) {
			int priority = 0;
			send(connectionInformation, message, priority);
		}

		/**
		 * @param connectionInformation
		 * @param message
		 * @param priority
		 */
		private void send(ConnectionInformation connectionInformation, P2PMessage message, int priority) {
			getPeerConnectionFor(connectionInformation).sendMessage(message);
		}

		public PeerConnection getPeerConnectionFor(ConnectionInformation connectionInformation) {
			PeerConnection conn;
			synchronized (peerConnections) {
				conn = peerConnections.get(connectionInformation);
			}
			if (conn == null) {
				conn = new WarnPeerConnection(connectionInformation);
			}
			return conn;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext#closeConnection(java.net.InetSocketAddress)
		 */
		public void closeConnection(ConnectionInformation peerAddress) {
			getPeerConnectionFor(peerAddress).close();
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext#openConnection(java.net.InetSocketAddress)
		 */
		public void openConnection(InetSocketAddress peerAddress) throws IOException {
			connectTo(peerAddress, new ConnectCallback() {

				public void connectFailed(InetSocketAddress remoteAddress, IOException reason) {
					
				}

				public void connectionOpen(PeerConnection conn) {
					
				}
				
			});
		}

	}

	public void onPeerMessage(ch.unibe.iam.mcast.core.PeerConnection connection, P2PMessage message) {
		onNewMessage(message, connection.getConnectionInformation());

	}

	public void onPeerConnectionClosed(ch.unibe.iam.mcast.core.PeerConnection connection) {
		connectionClosing(connection.getConnectionInformation());

	}

	public void onPeerConnectFailed(InetSocketAddress connection, IOException reason) {
		log.warn("failed connecting to peer " + connection, reason);
	}

	public MulticastCore getCore() {
		return core;
	}

	public boolean isConnectedTo(InetSocketAddress remote) {
		synchronized (peerConnections) {
			for (ConnectionInformation ci : peerConnections.keySet()) {
				if (ci.getRemoteAddress().equals(remote)) {
					return true;
				}
			}
		}
		return false;
	}

	private static class WarnPeerConnection extends PeerConnection {

		public WarnPeerConnection(ConnectionInformation peerAddress) {
			super(peerAddress.getLocalAddress(), peerAddress.getRemoteAddress());
		}

		@Override
		public void close() {
			log.warn("trying to close non-existing connection to " + super.getConnectionInformation());
		}

		@Override
		public void sendMessage(P2PMessage message) {
			log.warn("trying to send message " + message + " to non-existing connection to " + super.getConnectionInformation());

		}

		public void sendIPV4Packet(IPV4Packet packet, int priority) {
			log.warn("trying to send IPV4Packet " + packet + " to non-existing connection to " + super.getConnectionInformation());
		}

	}

}