/*
 * Created on Nov 11, 2004
 * 
 * TODO To change the template for this generated file go to Window - Preferences - Java - Code Style - Code Templates
 */
package ch.unibe.iam.mcast.router.p2p.filter;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.router.p2p.ConnectionInformation;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilter;
import ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext;
import ch.unibe.iam.mcast.router.p2p.message.KeepAliveMessage;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;

public class KeepAliveFilter implements P2PMessageFilter, Runnable {

    private static final Log log = LogFactory.getLog(KeepAliveFilter.class);

    private static final long KA_TIMEOUT = 15000;

    private final Set<ConnectionInformation> activePeers = new HashSet<ConnectionInformation>();

    private final Map<ConnectionInformation,Long> lastKeepAliveTimes = new HashMap<ConnectionInformation,Long>();

    private P2PMessageFilterContext context;

    private Thread thread;

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#handleMessage(java.net.InetSocketAddress, ch.unibe.iam.mcast.router.p2p.message.P2PMessage)
     */
    public synchronized boolean handleMessage(ConnectionInformation peerAddress, P2PMessage message) {
       
        if (message instanceof KeepAliveMessage) {
            touchPeer(peerAddress);
            return true;
        } else {
            return false;
        }
    }

    /**
     * @param peerAddress
     */
    private void touchPeer(ConnectionInformation peerAddress) {
        lastKeepAliveTimes.put(peerAddress, new Long(System.currentTimeMillis()));
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#onNewPeerConnection(java.net.InetSocketAddress)
     */
    public synchronized void onNewPeerConnection(ConnectionInformation peerAddress) {
        activePeers.add(peerAddress);
        touchPeer(peerAddress);
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#onPeerConnectionClosed(java.net.InetSocketAddress)
     */
    public synchronized void onPeerConnectionClosed(ConnectionInformation peerAddress) {
        if (log.isDebugEnabled())
            log.debug("connection closed to " + peerAddress);
        lastKeepAliveTimes.remove(peerAddress);
        activePeers.remove(peerAddress);
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#setContext(ch.unibe.iam.mcast.router.p2p.P2PMessageFilterContext)
     */
    public void setContext(P2PMessageFilterContext context) {
        this.context = context;
    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#start()
     */
    public synchronized void start() {
        if (log.isDebugEnabled())
            log.debug("KA filter started");
        this.thread = new Thread(this, "KeepAliveTimer");
        thread.start();

    }

    /*
     * (non-Javadoc)
     * 
     * @see ch.unibe.iam.mcast.router.p2p.P2PMessageFilter#stop()
     */
    public synchronized void stop() {
        if (log.isDebugEnabled())
            log.debug("KA filter stopped");
        this.thread.interrupt();
        try {
            this.thread.join();
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Runnable#run()
     */
public void run() {
        Random r = new Random();
        try {
            while (!Thread.interrupted()) {

                long sleepTimeout = (long) ((KA_TIMEOUT / 2 * r.nextDouble()) + KA_TIMEOUT / 4);
                Thread.sleep(sleepTimeout);

                long currentTime = System.currentTimeMillis();
                Set<ConnectionInformation> toClose=new HashSet<ConnectionInformation>();
                synchronized (this) {
                    KeepAliveMessage ka = new KeepAliveMessage();
                    for (ConnectionInformation peerAddress : activePeers) {                        
                        if (log.isDebugEnabled())
                            log.debug("sending KA to " + peerAddress);
                        context.sendMessage(peerAddress, ka);
                    }
                    

                    for (Map.Entry<ConnectionInformation, Long> entries : lastKeepAliveTimes.entrySet()) {
                    	ConnectionInformation peerAddress = entries.getKey();
                        long last=entries.getValue();
                        if ((currentTime - last) > KA_TIMEOUT) {
                            if (log.isDebugEnabled()) {

                                log.debug("did not receive KeepAlive from peer " + peerAddress + ". Closing connection");
                            }
                            toClose.add(peerAddress);
                            
                        }
                    }
                    
                }
                
                for (ConnectionInformation ci : toClose) {                        
                    context.closeConnection(ci);
                }

            }
        } catch (InterruptedException e) {
            // ignore and shutdown
        }

    }
}