/*
 * Created on Nov 8, 2004
 *
 * Object format:
 *  0                   1                   2                   3
 *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |     Class     |   Type        |  Length                       |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                                                               |
 * |                       Data                                    |
 * |                                                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 * Class: 	class of the object
 * Type:	type of the object (within the object class)
 * Length: 	Total length of the object in bytes (including the object header)
 * Data: 	the data of the object
 */
package ch.unibe.iam.mcast.router.p2p.message;

import java.io.UnsupportedEncodingException;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.net.ip.IPV4Packet;
import ch.unibe.iam.mcast.net.ip.MalformedPacketException;

/**
 * @author root
 * 
 * TODO To change the template for this generated type comment go to Window - Preferences - Java - Code Style - Code Templates
 */
public class P2PObjectUtil {

    /**
     * 
     */
    private P2PObjectUtil() {
        super();

    }

    /**
     * @param buffer
     * @param time
     */
    public static void serializeMsecLongTimestamp(ByteBuffer buffer, long time) {
        writeObjectHeader(buffer, P2PObjectConstants.CLASS_TIMESTAMP, P2PObjectConstants.TYPE_LONG_MSEC, 8);
        buffer.putLong(time);
    }

    public static void serializeIpPacketData(ByteBuffer buffer, IPV4Packet packet) {
        int length = packet.getPacketLength();
        writeObjectHeader(buffer, P2PObjectConstants.CLASS_DATA, P2PObjectConstants.TYPE_IP_PACKET_DATA, length);
        packet.serializePacket(buffer);
    }

    public static IPV4Packet deserializeIpPacketData(ByteBuffer buffer) throws MalformedP2PMessageException {
        int dataLength = checkClassAndType(buffer, P2PObjectConstants.CLASS_DATA, P2PObjectConstants.TYPE_IP_PACKET_DATA);
        try {
            return IPV4Packet.createIPV4Packet(buffer);
        } catch (MalformedPacketException e) {
            throw new MalformedP2PMessageException(e);
        }

    }

    public static void serializeShortPortNumber(short portNumber, ByteBuffer buffer) {
        int dataLength = 2;
        writeObjectHeader(buffer, P2PObjectConstants.CLASS_PORT_NUMBER, P2PObjectConstants.TYPE_SHORT_PORT_NUMBER, dataLength);
        buffer.putShort(portNumber);
    }

    /**
     * @param buffer
     * @param clazz
     * @param type
     * @param dataLength
     */
    public static void writeObjectHeader(ByteBuffer buffer, byte clazz, byte type, int dataLength) {
        buffer.put(clazz);

        buffer.put(type);
        buffer.putShort((short) (4 + dataLength));
    }

    
    public static void serializeString(String value, ByteBuffer buffer) {
    	int length=value.length()*2;
    	writeObjectHeader(buffer, P2PObjectConstants.CLASS_STRING, P2PObjectConstants.TYPE_UNICODE16, length);
    	for (int i=0; i<value.length(); i++) {
    		buffer.putChar(value.charAt(i));
    	}    	
    }

    public static String deserializeString(ByteBuffer buffer) throws MalformedP2PMessageException {
    	int len=checkClassAndType(buffer, P2PObjectConstants.CLASS_STRING, P2PObjectConstants.TYPE_UNICODE16);
    	byte data[]=new byte[len];
    	buffer.get(data);
    	String ret;
		try {
			ret = new String(data,"UTF-16");
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException("this should never happend",e);
		}
    	return ret;
    }

    
    public static long deserializeMsecLongTimestamp(ByteBuffer buffer) throws MalformedP2PMessageException {
        checkClassTypeAndDataLength(buffer, P2PObjectConstants.CLASS_TIMESTAMP, P2PObjectConstants.TYPE_LONG_MSEC, 8);
        return buffer.getLong();
    }

    public static short deserializeShortPortNumber(ByteBuffer buffer) throws MalformedP2PMessageException {
        checkClassTypeAndDataLength(buffer, P2PObjectConstants.CLASS_PORT_NUMBER, P2PObjectConstants.TYPE_SHORT_PORT_NUMBER, 2);

        return buffer.getShort();
    }

    public static void serializeInetSocketAddress(InetSocketAddress address, ByteBuffer buffer) {
        byte type = address.getAddress() instanceof Inet6Address ? P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V6 : P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V4;
        short objectLength = (short) (4 + (type == P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V6 ? 16 : 4));
        writeObjectHeader(buffer, P2PObjectConstants.CLASS_INET_SOCKET_ADDRESS, type, objectLength);
        buffer.putShort(objectLength);
        buffer.put(address.getAddress().getAddress());
        buffer.putInt(address.getPort());
    }

     
    
    public static void serializeInetAddress(InetAddress address, ByteBuffer buffer) {
        byte objectClass = P2PObjectConstants.CLASS_INET_ADDRESS;

        byte type = address instanceof Inet6Address ? P2PObjectConstants.TYPE_INET_ADDRES_V6 : P2PObjectConstants.TYPE_INET_ADDRES_V4;
        int objectLength = (type == P2PObjectConstants.TYPE_INET_ADDRES_V6 ? 16 : 4);

        writeObjectHeader(buffer, objectClass, type, objectLength);
        buffer.put(address.getAddress());
    }

    /**
//     * @param buffer
//     * @param objectClass
//     * @param type
//     * @param objectLength
//     */
//    private static void putObjectHeader(ByteBuffer buffer, byte objectClass, byte type, int objectLength) {
//        buffer.put(objectClass);
//        buffer.put(type);
//
//        buffer.putShort((short) objectLength);
//    }

    public static InetAddress deserializeInetAddress(ByteBuffer buffer) throws MalformedP2PMessageException {

        byte type = parseTypeWithExpectedClass(buffer, P2PObjectConstants.CLASS_INET_ADDRESS);

        if (type != P2PObjectConstants.TYPE_INET_ADDRES_V4 && type != P2PObjectConstants.TYPE_INET_ADDRES_V6) {
            throw new MalformedP2PMessageException("unknown object type " + type + " for class " + P2PObjectConstants.CLASS_INET_ADDRESS);
        }

        int length = buffer.getShort() & 0xffff;

        if (buffer.remaining() < length - 4) {
            throw new MalformedP2PMessageException("object bigger than buffer");
        }

        int addressLength = type == P2PObjectConstants.TYPE_INET_ADDRES_V6 ? 16 : 4;

        if (addressLength + 4 != length) {
            throw new MalformedP2PMessageException("object has wrong size (is " + length + ", should be " + (addressLength + 4) + ")");
        }

        byte[] addressBuffer = new byte[addressLength];

        int oldLimit = buffer.limit();

        buffer.limit(buffer.position() + length - 4);
        buffer.get(addressBuffer);
        InetAddress address;
        try {
            address = InetAddress.getByAddress(addressBuffer);
        } catch (UnknownHostException e) {
            throw new RuntimeException("this should never happend", e);
        }
        
        buffer.limit(oldLimit);

        return address;
    }

    public static InetSocketAddress deserializeInetSocketAddress(ByteBuffer buffer) throws MalformedP2PMessageException {

        byte expectedClass = P2PObjectConstants.CLASS_INET_SOCKET_ADDRESS;

        byte type = parseTypeWithExpectedClass(buffer, expectedClass);

        if (type != P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V4 && type != P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V6) {
            throw new MalformedP2PMessageException("unknown object type " + type + " for class " + expectedClass);
        }

        int length = buffer.getShort() & 0xffff;

        if (buffer.remaining() < length - 4) {
            throw new MalformedP2PMessageException("object bigger than buffer");
        }
        int addressLength = type == P2PObjectConstants.TYPE_INET_SOCKET_ADDRES_V6 ? 16 : 4;
        if (addressLength + 8 != length) {
            throw new MalformedP2PMessageException("object has wrong size (is " + length + ", should be " + (addressLength + 8) + ")");
        }

        byte[] addressBuffer = new byte[addressLength];

        int startLimit = buffer.limit();

        buffer.limit(buffer.position() + length - 4);
        buffer.get(addressBuffer);
        InetAddress address;
        try {
            address = InetAddress.getByAddress(addressBuffer);
        } catch (UnknownHostException e) {
            throw new RuntimeException("this should never happend", e);
        }

        int port = buffer.getInt();
        buffer.limit(startLimit);
        return new InetSocketAddress(address, port);
    }

    /**
     * @param buffer
     * @param expectedClass
     * @return
     * @throws MalformedP2PMessageException
     */
    private static byte parseTypeWithExpectedClass(ByteBuffer buffer, byte expectedClass) throws MalformedP2PMessageException {
        if (buffer.remaining() < 4) {
            throw new MalformedP2PMessageException("object must have at least 4 bytes");
        }

        byte cls = buffer.get();
        if (cls != expectedClass) {
            throw new MalformedP2PMessageException("expected object of clas " + (expectedClass & 0xff) + " but got object of class " + (cls & 0xff));
        }

        byte type = buffer.get();
        return type;
    }

    private static void checkClassTypeAndDataLength(ByteBuffer buffer, byte expectedClass, byte expectedType, int dataLength) throws MalformedP2PMessageException {
        int objectLength = checkClassAndType(buffer, expectedClass, expectedType);

        if (objectLength != dataLength) {
            throw new MalformedP2PMessageException("got message with data length" + (objectLength) + " expected was " + dataLength);
        }
        if (buffer.remaining() < dataLength) {
            throw new MalformedP2PMessageException("expected data of length " + dataLength + " but the remaining buffer contains " + buffer.remaining() + " bytes ");
        }
    }

    /**
     * @param buffer
     * @param expectedClass
     * @param expectedType
     * @throws MalformedP2PMessageException
     */
    private static int checkClassAndType(ByteBuffer buffer, byte expectedClass, byte expectedType) throws MalformedP2PMessageException {
        if (buffer.remaining() < 4) {
            throw new MalformedP2PMessageException("object must have at least 4 bytes");
        }

        byte cls = buffer.get();
        if (cls != expectedClass) {
            throw new MalformedP2PMessageException("expected object of clas " + (expectedClass & 0xff) + " but got object of class " + (cls & 0xff));
        }

        byte type = buffer.get();
        if (type != expectedType) {
            throw new MalformedP2PMessageException("expected object of type " + (expectedType & 0xff) + " but got object of type " + (expectedType & 0xff));
        }
        return (buffer.getShort() & 0xffff) - 4;

    }

    public static int deserializeQosRoutePriority(ByteBuffer buffer) throws MalformedP2PMessageException {
        checkClassTypeAndDataLength(buffer, P2PObjectConstants.CLASS_QOS, P2PObjectConstants.TYPE_ROUTE_PRIORITY, 4);
        return buffer.getInt();
    }

    public static void serializeQoSRoutePriority(ByteBuffer buffer, int priority) {
        writeObjectHeader(buffer,P2PObjectConstants.CLASS_QOS, P2PObjectConstants.TYPE_ROUTE_PRIORITY,4);
        buffer.putInt(priority);
    }

}