package ch.unibe.iam.mcast.router.p2p.message.dynaddr;

import java.net.Inet6Address;
import java.net.InetAddress;
import java.nio.ByteBuffer;

import ch.unibe.iam.mcast.router.p2p.message.MalformedP2PMessageException;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessage;
import ch.unibe.iam.mcast.router.p2p.message.P2PMessageConstants;
import ch.unibe.iam.mcast.router.p2p.message.P2PObjectUtil;

public class DynamicAddressResponseMessage extends P2PMessage {
    
    private InetAddress interfaceAddress;

    public DynamicAddressResponseMessage() {
        super(P2PMessageConstants.DYNAMIC_ADDRESS_RESPONSE_MESSAGE);
    }
    public DynamicAddressResponseMessage(InetAddress address) {
        this();
        this.interfaceAddress=address;
    }

    protected void serializeContent(ByteBuffer buffer) {
        if (interfaceAddress == null) {
            throw new RuntimeException("trying to serialize null iface address");
        }
        P2PObjectUtil.serializeInetAddress(interfaceAddress,buffer);       
    }

    protected int contentLength() {     
        return interfaceAddress instanceof Inet6Address ? 16+4 : 4+4;
    }

    protected void deserializeContent(ByteBuffer buffer) throws MalformedP2PMessageException {
        
        interfaceAddress=P2PObjectUtil.deserializeInetAddress(buffer);
    }
    public InetAddress getInterfaceAddress() {
        return interfaceAddress;
    }

}
