package ch.unibe.iam.mcast.router.pastry;

import java.io.Serializable;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import rice.Continuation;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.IdRange;
import rice.p2p.commonapi.IdSet;
import rice.persistence.Storage;

public class DumbStorage implements Storage {

	Map<Id, Serializable> values;

	Map<Id, Serializable> metaData;

	public synchronized void store(Id id, Serializable metadata, Serializable obj, Continuation c) {
		values.put(id, obj);
		metaData.put(id, metadata);
		c.receiveResult(Boolean.TRUE);

	}

	public synchronized void unstore(Id id, Continuation c) {

		if (values.containsKey(id)) {
			values.remove(id);
			metaData.remove(id);
			c.receiveResult(Boolean.TRUE);

		} else {
			c.receiveResult(Boolean.FALSE);
		}

	}

	public synchronized boolean exists(Id id) {
		return values.containsKey(id);
	}

	public synchronized void flush(Continuation c) {
		values.clear();
		metaData.clear();
		c.receiveResult(Boolean.TRUE);

	}

	public synchronized Serializable getMetadata(Id id) {
		return metaData.get(id);
	}

	public synchronized void getObject(Id id, Continuation c) {
		c.receiveResult(values.get(id));
	}

	public synchronized int getSize() {
		return values.size();
	}

	public synchronized long getTotalSize() {
		return getSize() * 5;
	}

	public synchronized void rename(Id oldId, Id newId, Continuation c) {
		Serializable v = values.remove(oldId);
		Serializable m = metaData.remove(oldId);
		values.put(newId, v);
		metaData.put(newId, m);
		c.receiveResult(Boolean.TRUE);

	}

	public synchronized IdSet scan() {
		rice.pastry.IdSet ret = new rice.pastry.IdSet();
		for (Id id : values.keySet()) {
			ret.addId(id);
		}
		return ret;
	}

	public synchronized IdSet scan(IdRange range) {
		rice.pastry.IdSet ret = new rice.pastry.IdSet();
		for (Id id : values.keySet()) {
			if (range.containsId(id)) {
				ret.addId(id);
			}
		}
		return ret;
	}

	public synchronized SortedMap scanMetadata() {
		TreeMap tm=new TreeMap(metaData);
		return tm;
	}

	public synchronized SortedMap scanMetadata(IdRange range) {
		TreeMap ret=new TreeMap();
		for (Map.Entry<Id, Serializable> me : metaData.entrySet()) {
			if (range.containsId(me.getKey())) {
				ret.put(me.getKey(),me.getValue());
			}
		}
		return ret;		
	}

	public synchronized SortedMap scanMetadataValuesHead(Object value) {
		Comparable comp=(Comparable) value;
		TreeMap ret=new TreeMap();
		for (Map.Entry<Id, Serializable> me : metaData.entrySet()) {
			if (comp.compareTo(me.getValue()) < 1 ) {
				ret.put(me.getKey(),me.getValue());
			}
		}
		return ret;		
	}

	public synchronized SortedMap scanMetadataValuesNull() {
		TreeMap ret=new TreeMap();
		for (Map.Entry<Id, Serializable> me : metaData.entrySet()) {
			if (me.getValue() == null) {
				ret.put(me.getKey(),me.getValue());
			}
		}
		return ret;		
	}

	public synchronized void setMetadata(Id id, Serializable metadata, Continuation command) {
		this.metaData.put(id, metadata);
		command.receiveResult(Boolean.TRUE);

	}

}
