package ch.unibe.iam.mcast.router.pastry;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeoutException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rice.Continuation;
import rice.environment.Environment;
import rice.p2p.commonapi.Application;
import rice.p2p.commonapi.Endpoint;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.NodeHandleSet;
import rice.p2p.commonapi.RouteMessage;
import rice.p2p.past.Past;
import rice.p2p.past.PastImpl;
import rice.pastry.NodeIdFactory;
import rice.pastry.PastryNode;
import rice.pastry.commonapi.PastryIdFactory;
import rice.pastry.socket.SocketPastryNodeFactory;
import rice.pastry.socket.nat.StubNATHandler;
import rice.pastry.socket.nat.sbbi.SBBINatHandler;
import rice.pastry.standard.RandomNodeIdFactory;
import rice.persistence.LRUCache;
import rice.persistence.MemoryStorage;
import rice.persistence.StorageManagerImpl;
import ch.unibe.iam.mcast.router.OverlayTopologyChangeListener;
import ch.unibe.iam.mcast.router.pastry.message.BackOffMessage;
import ch.unibe.iam.mcast.router.pastry.message.BootstrapMessage;
import ch.unibe.iam.mcast.router.pastry.message.DHTMessage;
import ch.unibe.iam.mcast.router.pastry.message.DHTResponseMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinFailedMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinFinishedMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinToMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinedMessage;
import ch.unibe.iam.mcast.router.pastry.message.LeftMessage;
import ch.unibe.iam.mcast.router.pastry.message.NewGroupStateMessage;
import ch.unibe.iam.mcast.router.pastry.message.WantToJoinMessage;
import ch.unibe.iam.mcast.router.pastry.past.ActiveRingContent;
import ch.unibe.iam.mcast.router.pastry.past.CreateNewRingContent;
import ch.unibe.iam.mcast.router.pastry.past.GroupBootstrapContent;
import ch.unibe.iam.mcast.router.pastry.past.LeavingGroupContent;
import ch.unibe.iam.mcast.router.pastry.past.OfferNewRingContent;
import ch.unibe.iam.mcast.router.pastry.past.PeerAliveContent;
import ch.unibe.iam.mcast.router.pastry.scribe.ScribeRing;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;

public class GroupPeerDHT {
	private static final String REPLICATE_DHT_ENV_KEY = "REPLICATE_DHT";

	private static final long WANT_TO_JOIN_TIMEOUT = 15000;

	private static final long BOOTSTRAP_BACKOFF_TIMEOUT = 3000;

	private static final Log log = LogFactory.getLog(GroupPeerDHT.class);

	private final InetSocketAddress bindAddress;

	// private Map<InetAddress, MessageFuture> futures = new
	// HashMap<InetAddress, MessageFuture>();
	//
	// private Map<InetAddress, GroupState> groupStates = new
	// HashMap<InetAddress, GroupState>();

	private Map<InetAddress, Future<ScribeRing, RuntimeException>> joinedGroups = new HashMap<InetAddress, Future<ScribeRing, RuntimeException>>();

	// private Map<InetAddress, Set<Id>> knownReplicas = new
	// HashMap<InetAddress, Set<Id>>();

	private final Environment environment = new Environment();

	private PastryNode node;

	// private Endpoint endpoint;

	private final OverlayTopologyChangeListener listener;

	private SocketPastryNodeFactory factory;

	private final InetSocketAddress dataSocketAddress;

	private final InetSocketAddress[] bootstrapNodes;

	private static final Latch operationLatch = new Latch();

	private static final int MAX_REPLICAS = 5;

	private Thread joinedDistributer;

	private Past pastApp;

	public GroupPeerDHT(InetSocketAddress bindAddress, InetSocketAddress[] bootstrapNodes, InetSocketAddress dataSocketAddress, OverlayTopologyChangeListener listener)
			throws IOException {
		this.bootstrapNodes = bootstrapNodes;
		environment.getParameters().setBoolean("pastry_factory_selectorPerNode", true); // set
		environment.getParameters().setBoolean("pastry_factory_processorPerNode", true); // set
		environment.getParameters().setBoolean("pastry_factory_multipleNodes", true); // set
		environment.getParameters().setString("nat_search_policy", "never");
		this.bindAddress = bindAddress;
		this.dataSocketAddress = dataSocketAddress;
		this.listener = listener;
	}

	public synchronized void start() throws InterruptedException, IOException {
		System.out.println("-> starting group peer DHT bind address: " + bindAddress);
		InetAddress host = bindAddress.getAddress();
		NodeIdFactory nidFactory = new RandomNodeIdFactory(environment);

		// factory = host == null ? new SocketPastryNodeFactory(nidFactory,
		// bindAddress.getPort(), environment) : new
		// SocketPastryNodeFactory(nidFactory, host, bindAddress.getPort(),
		// environment, new
		// SBBINatHandler(environment,bindAddress.getAddress()));

		factory = host == null ? new SocketPastryNodeFactory(nidFactory, bindAddress.getPort(), environment) : new SocketPastryNodeFactory(nidFactory, host, bindAddress.getPort(),
				environment, new StubNATHandler(environment, bindAddress.getAddress()));

		// This will return null if we there is no node at that location
		NodeHandle bootHandle = bootstrapNodes == null ? null : factory.getNodeHandle(bootstrapNodes);

		System.out.println(">>>> boot handle: " + bootHandle);

		// construct a node, passing the null boothandle on the first loop will
		// cause the node to start its own ring
		node = factory.newNode((rice.pastry.NodeHandle) bootHandle);

		// the node may require sending several messages to fully boot into the
		// ring
		while (!node.isReady()) {
			// delay so we don't busy-wait
			Thread.sleep(100);
		}
		System.out.println("node ready !");

		PastryIdFactory idf = new rice.pastry.commonapi.PastryIdFactory(environment);

		DummyMemoryStorage ms = new DummyMemoryStorage(idf);

		this.pastApp = new PastImpl(node, new StorageManagerImpl(idf, ms, new LRUCache(new DummyMemoryStorage(idf), 30, environment)), 5, "boo");

		// this.endpoint = node.buildEndpoint(this, "GruoupPeerDHT");
		// this.endpoint.register();
		//
		this.joinedDistributer = new Thread(new JoinedMessageDistributer());
		this.joinedDistributer.start();
		//
		// // joinTo(InetAddress.getByAddress(new byte[] { (byte) 224, 0, 0, 1
		// }));
		operationLatch.openLatch();
	}

	public synchronized void stop() {
		environment.destroy();
		operationLatch.close();
		if (joinedDistributer != null)
			this.joinedDistributer.interrupt();
	}

	// public void deliver(Id destinationId, Message msg) {
	//
	// // NodeHandleSet nhs=;
	// // System.out.println("node handle set size: "+nhs.size());
	//
	// boolean hot_server = msg instanceof DHTMessage ?
	// endpoint.localLookup(generateNodeId(((DHTMessage) msg).getGroup()), 5,
	// false).size() == 0 : true;
	//
	// // if (!hot_server) {
	// // System.out.println("received replica "+msg);
	// // }
	//
	// if (!(msg instanceof JoinedMessage)) {
	// System.out.println("received " + msg + " for " + destinationId +" hot:
	// "+hot_server);
	// }
	// // System.out.println("forwarding to " +
	// // routeMessage.getDestinationId());
	// // return true;
	// if (msg instanceof DHTResponseMessage) {
	// // TODO match query/response with IDs!
	// DHTResponseMessage resp = (DHTResponseMessage) msg;
	// synchronized (futures) {
	// MessageFuture future = futures.remove(resp.getGroup());
	// if (future != null) {
	// future.set(msg);
	// }
	// }
	// } else if (msg instanceof WantToJoinMessage) {
	// WantToJoinMessage wtj = (WantToJoinMessage) msg;
	// replicateMessage(wtj, hot_server);
	// synchronized (groupStates) {
	// if (!groupStates.containsKey(wtj.getGroup())) {
	// sendMessageIfHot(wtj.getSender(), new BootstrapMessage(wtj.getGroup()),
	// hot_server);
	// groupStates.put(wtj.getGroup(), new GroupState(wtj.getGroup(),
	// wtj.getSender()));
	// } else {
	// GroupState state = groupStates.get(wtj.getGroup());
	// if (state.getCurrentState() == GroupState.State.ACTIVE) {
	// sendMessageIfHot(wtj.getSender(), new JoinToMessage(wtj.getGroup(),
	// state.getActivePeers()), hot_server);
	// } else if (state.getCurrentState() == GroupState.State.BOOTSTRAPPING) {
	// sendMessageIfHot(wtj.getSender(), new BackOffMessage(wtj.getGroup(),
	// BOOTSTRAP_BACKOFF_TIMEOUT), hot_server);
	// } else {
	// log.fatal("unknown group state" + state.getCurrentState());
	// }
	// }
	// }
	//
	// } else if (msg instanceof JoinedMessage) {
	//
	// JoinedMessage joined = (JoinedMessage) msg;
	// replicateMessage(joined, hot_server);
	// synchronized (groupStates) {
	// if (!groupStates.containsKey(joined.getGroup())) {
	// GroupState groupState = new GroupState(joined.getGroup(),
	// joined.getSender());
	// groupState.addActivePeer(joined.getEndpoint());
	// groupStates.put(joined.getGroup(), groupState);
	// } else {
	// GroupState groupState = groupStates.get(joined.getGroup());
	// GroupState.State oldState = groupState.getCurrentState();
	// groupState.addActivePeer(joined.getEndpoint());
	// if (oldState == GroupState.State.BOOTSTRAPPING) {
	// sendMessageIfHot(joined.getSender(), new
	// JoinFinishedMessage(joined.getGroup()), hot_server);
	// }
	// }
	// }
	// } else if (msg instanceof LeftMessage) {
	//
	// LeftMessage left = (LeftMessage) msg;
	// replicateMessage(left, hot_server);
	// synchronized (groupStates) {
	// if (!groupStates.containsKey(left.getGroup())) {
	// log.warn("removed left message for unknown group" + left.getGroup());
	// } else {
	// GroupState groupState = groupStates.get(left.getGroup());
	// if (groupState.removeActivePeer(left.getEndpoint())) {
	// groupStates.remove(left.getEndpoint());
	// }
	// }
	// }
	//
	// } else if (msg instanceof JoinFailedMessage) {
	// JoinFailedMessage failed = (JoinFailedMessage) msg;
	// replicateMessage(failed, hot_server);
	// synchronized (groupStates) {
	// if (!groupStates.containsKey(failed.getGroup())) {
	// sendMessageIfHot(failed.getSender(), new
	// BootstrapMessage(failed.getGroup()), hot_server);
	// groupStates.put(failed.getGroup(), new GroupState(failed.getGroup(),
	// failed.getSender()));
	// } else {
	// groupStates.remove(failed.getGroup());
	// GroupState groupState = new GroupState(failed.getGroup(),
	// failed.getSender());
	// groupStates.put(failed.getGroup(), groupState);
	// sendMessageIfHot(failed.getSender(), new
	// BootstrapMessage(failed.getGroup()), hot_server);
	// }
	// }
	// } else if (msg instanceof NewGroupStateMessage) {
	// NewGroupStateMessage ngs = (NewGroupStateMessage) msg;
	// System.out.println(">>>>>>>>>>>>>>>> got new group state for " +
	// ngs.getGroup());
	// synchronized (groupStates) {
	// if (ngs.getNewState() == null) {
	// groupStates.remove(ngs.getGroup());
	// } else {
	// groupStates.put(ngs.getGroup(), ngs.getNewState());
	// }
	// }
	// } else {
	//
	// log.warn("received messages of unknown type " + msg);
	// }
	// }
	//
	// public void sendMessageIfHot(Id destination, DHTMessage msg, boolean
	// hot_server) {
	// if (hot_server) {
	// sendMessage(destination, msg);
	// }
	// }
	//
	// public void replicateMessage(DHTMessage msg, boolean hot_server) {
	// if (hot_server &&
	// Boolean.parseBoolean(System.getenv(REPLICATE_DHT_ENV_KEY))) {
	//
	// InetAddress group = msg.getGroup();
	// NodeHandleSet nhs = endpoint.replicaSet(generateNodeId(group),
	// MAX_REPLICAS);
	//
	// synchronized (knownReplicas) {
	// synchronized (groupStates) {
	// GroupState state = groupStates.get(group);
	// Set<Id> replicas = knownReplicas.get(group);
	// if (replicas == null) {
	// replicas = new HashSet<Id>();
	// }
	//
	// for (int i = 0; i < nhs.size(); i++) {
	// NodeHandle hl = nhs.getHandle(i);
	// Id handleId = hl.getId();
	// if (!replicas.contains(handleId)) {
	// System.out.println("***** ow ow ow new replica" + handleId);
	// replicas.add(handleId);
	// if (!endpoint.getId().equals(handleId)) {
	// System.out.println("***** sending state to the new replica" + handleId);
	// sendMessage(handleId, new NewGroupStateMessage(group, state));
	// }
	//
	// }
	// }
	//
	// for (Iterator<Id> it=replicas.iterator(); it.hasNext();) {
	// Id replica = it.next();
	// if (nhs.getHandle(replica) == null) {
	// it.remove();
	// }
	// }
	//
	// if (replicas.size() == 0) {
	// knownReplicas.remove(group);
	// } else {
	// if (!knownReplicas.containsKey(group)) {
	// knownReplicas.put(group, replicas);
	// }
	// }
	//
	// }
	//
	// }
	//
	// // System.out.println("replicating "+msg+" to replicas
	// // "+nhs.size());
	// for (int i = 0; i < nhs.size(); i++) {
	// if (!nhs.getHandle(i).getId().equals(endpoint.getId()))
	// sendMessage(nhs.getHandle(i).getId(), msg);
	// }
	//
	// }
	// }

	public ScribeRing joinTo(final InetAddress group, final QoSRequirement qos) throws InterruptedException {

		operationLatch.pass();

		final Future<ScribeRing, RuntimeException> retFuture;

		boolean running = false;

		synchronized (joinedGroups) {
			if (joinedGroups.containsKey(group)) {
				log.warn("already joined");
				retFuture = joinedGroups.get(group);
				running = true;
			} else {
				retFuture = new Future<ScribeRing, RuntimeException>();
				joinedGroups.put(group, retFuture);
			}
		}

		if (!running) {

			Thread t = new Thread(new Runnable() {

				public void run() {

					Id myId = pastApp.getLocalNodeHandle().getId();
					Id groupId = generateNodeId(group);

					for (;;) {

						System.out.println("offering new ring content");

						OfferNewRingContent offer = new OfferNewRingContent(groupId, myId);
						final Latch offerLatch = new Latch();
						pastApp.insert(offer, new Continuation() {

							public void receiveException(Exception result) {
								offerLatch.openLatchWithException(result);

							}

							public void receiveResult(Object result) {
								offerLatch.openLatch();

							}

						});

						try {
							offerLatch.pass();
						} catch (InterruptedException e2) {
							throw new RuntimeException(e2);
						}

						System.out.println("offering new ring content finished.");

						System.out.println("lookup..");
						final Future<GroupBootstrapContent, Exception> stateFuture = new Future<GroupBootstrapContent, Exception>();

						pastApp.lookup(groupId, new Continuation() {

							public void receiveException(Exception result) {
								stateFuture.setException(result);

							}

							public void receiveResult(Object result) {
								System.out.println("++++ result: " + result);
								stateFuture.set((GroupBootstrapContent) result);

							}
						});

						GroupBootstrapContent ret;
						try {
							ret = stateFuture.get(25000);
							System.out.println("result: " + ret);
						} catch (TimeoutException e1) {
							throw new RuntimeException(e1);
						} catch (Exception e1) {
							throw new RuntimeException(e1);
						}

						if (ret instanceof CreateNewRingContent) {
							CreateNewRingContent cnrc = (CreateNewRingContent) ret;
							if (cnrc.getBootstrapId().equals(myId)) {
								ScribeRing ring;
								try {
									ring = new ScribeRing(qos, null, factory, dataSocketAddress, group, listener);
								} catch (IOException e) {
									throw new RuntimeException(e);
								} catch (InterruptedException e) {
									throw new RuntimeException(e);
								}

								ActiveRingContent arc = new ActiveRingContent(groupId, myId);
								arc.refresh(ring.getLocalAddress());

								pastApp.insert(arc, new Continuation() {

									public void receiveException(Exception result) {
										// TODO Auto-generated method stub

									}

									public void receiveResult(Object result) {
										// TODO Auto-generated method stub

									}

								});

								retFuture.set(ring);
								return;
							}
						} else if (ret instanceof ActiveRingContent) {
							ActiveRingContent arc = (ActiveRingContent) ret;
							ScribeRing ring;
							try {
								try {
									ring = new ScribeRing(qos, arc.getBootstrapEntries().toArray(new InetSocketAddress[0]), factory, dataSocketAddress, group, listener);
								} catch (InterruptedException e) {
									throw new RuntimeException(e);
								}
							} catch (IOException e) {
								throw new RuntimeException(e);
							}

							retFuture.set(ring);
							return;
						}

						try {
							Thread.sleep(1000);
						} catch (InterruptedException e) {
							throw new RuntimeException(e);
						}
					}
				}
			});

			t.start();

		}

		try {
			return retFuture.get(Long.MAX_VALUE);
		} catch (TimeoutException e) {
			throw new RuntimeException("this should never happend", e);
		}
	}

	public void leaveGroup(InetAddress group) {
		try {
			operationLatch.pass();
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
		
		System.out.println("!!!!! leaving group "+group);
		ScribeRing ring = null;
		synchronized (joinedGroups) {
			

			Future<ScribeRing, RuntimeException> removed = joinedGroups.remove(group);
			if (removed == null) {
				log.warn("was not joined to group " + group);
				return;
			} else {
				try {
					ring = removed.get(Long.MAX_VALUE);
				} catch (TimeoutException e) {
					throw new RuntimeException(e);
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}
			}

			System.out.println("!!!!! leaving group "+group+", ring: "+ring);
			
			Id myId = pastApp.getLocalNodeHandle().getId();
			pastApp.insert(new LeavingGroupContent(generateNodeId(group), myId, ring.getLocalAddress()), new Continuation() {

				public void receiveException(Exception result) {
					log.warn(result);

				}

				public void receiveResult(Object result) {

				}

			});
			
			System.out.println("left group "+group);
		}
		// sendMessage(group, new LeftMessage(myId, group,
		// ring.getLocalAddress()));

		ring.stop();
	}

	// private Message sendAndWait(InetAddress group, Message message) throws
	// InterruptedException {
	// Id destinationId = generateNodeId(group);
	// MessageFuture future = new MessageFuture();
	// synchronized (futures) {
	// futures.put(group, future);
	// }
	//
	// endpoint.route(destinationId, message, null);
	// boolean received = false;
	//
	// Message resp = null;
	//
	// while (!received) {
	// try {
	// resp = future.get(WANT_TO_JOIN_TIMEOUT);
	// synchronized (futures) {
	// futures.remove(group);
	// }
	// received = true;
	// } catch (TimeoutException e) {
	// endpoint.route(destinationId, message, null);
	// }
	// }
	// return resp;
	// }

	// private void sendMessage(Id node, Message message) {
	// endpoint.route(node, message, null);
	// // endpoint.networkNeighbors(num);
	// }

	// private void sendMessage(InetAddress group, Message message) {
	// sendMessage(generateNodeId(group), message);
	// }

	// public boolean forward(RouteMessage routeMessage) {
	// // System.out.println("forwarding " + routeMessage);
	// // System.out.println("forwarding to " +
	// // routeMessage.getDestinationId().toStringFull());
	// return true;
	// }

	// public void update(NodeHandle arg0, boolean arg1) {
	// // TODO Auto-generated method stub
	//
	// }

	// public static void main(String argv[]) throws InterruptedException,
	// IOException {
	// GroupPeerDHT dht1 = new GroupPeerDHT(new
	// InetSocketAddress(InetAddress.getLocalHost(), 1025), null,new
	// InetSocketAddress(InetAddress.getLocalHost(), 2222), new
	// OverlayTopologyChangeListener(){
	//
	// public void addConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("+ connect to "+connection+" for "+group);
	//				
	// }
	//
	// public void removeConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("- disconnect to "+connection+" for "+group);
	//				
	// }});
	//		
	// GroupPeerDHT dht3 = new GroupPeerDHT(new
	// InetSocketAddress(InetAddress.getLocalHost(), 2025), new
	// InetSocketAddress(InetAddress.getLocalHost(), 1025),new
	// InetSocketAddress(InetAddress.getLocalHost(), 2223), new
	// OverlayTopologyChangeListener(){
	//
	// public void addConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("+ connect to "+connection+" for "+group);
	//				
	// }
	//
	// public void removeConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("- disconnect to "+connection+" for "+group);
	//				
	// }});
	//		
	//		
	// GroupPeerDHT dht2 = new GroupPeerDHT(new
	// InetSocketAddress(InetAddress.getLocalHost(), 3025), new
	// InetSocketAddress(InetAddress.getLocalHost(), 1025),new
	// InetSocketAddress(InetAddress.getLocalHost(), 2224), new
	// OverlayTopologyChangeListener(){
	//
	// public void addConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("+ connect to "+connection+" for "+group);
	//				
	// }
	//
	// public void removeConnection(InetAddress group, InetSocketAddress
	// connection) {
	// System.out.println("- disconnect to "+connection+" for "+group);
	//				
	// }});
	// dht1.start();
	// dht2.start();
	// dht3.start();
	// InetAddress group = InetAddress.getByAddress(new byte[] { (byte) 224, 0,
	// 0, 1 });
	// dht1.joinTo(group,QoSRequirement.BEST_EFFORT);
	// //Thread.sleep(2000);
	// dht2.joinTo(group,QoSRequirement.BEST_EFFORT);
	// //Thread.sleep(4000);
	// dht3.joinTo(group,QoSRequirement.BEST_EFFORT);
	// Thread.sleep(34000);
	// System.out.println("leaving ...");
	// dht2.leaveGroup(group);
	// Thread.sleep(15000);
	// System.out.println("joining ...");
	// dht2.joinTo(group, QoSRequirement.BEST_EFFORT);
	// }

	public Id generateNodeId(InetAddress groupAddress) {
		byte rawIP[] = groupAddress.getAddress();
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA");
		} catch (NoSuchAlgorithmException e) {
			log.fatal(e);
		}

		md.update(rawIP);
		byte[] digest = md.digest();

		Id nodeId = rice.pastry.Id.build(digest);

		return nodeId;
	}

	private class JoinedMessageDistributer implements Runnable {

		public void run() {
			Id myId = pastApp.getLocalNodeHandle().getId();
			while (!Thread.interrupted()) {
				Set<PeerAliveContent> messages = new HashSet<PeerAliveContent>();

				synchronized (joinedGroups) {
					for (InetAddress group : joinedGroups.keySet()) {
						PeerAliveContent pac;
						try {
							pac = new PeerAliveContent(generateNodeId(group), myId, joinedGroups.get(group).get(Long.MAX_VALUE).getLocalAddress());
						} catch (TimeoutException e) {
							throw new RuntimeException(e);
						} catch (InterruptedException e) {
							throw new RuntimeException(e);
						}
						// JoinedMessage msg = new
						// JoinedMessage(endpoint.getId(), group,
						// joinedGroups.get(group).getLocalAddress());
						messages.add(pac);
					}

				}

				for (PeerAliveContent pac : messages) {

					pastApp.insert(pac, new Continuation() {

						public void receiveException(Exception result) {
							// TODO Auto-generated method stub

						}

						public void receiveResult(Object result) {
							// TODO Auto-generated method stub

						}

					});

				}

				try {
					Thread.sleep(2000);
				} catch (InterruptedException e) {
					return;
				}
			}

		}

	}

}
