package ch.unibe.iam.mcast.router.pastry;

import java.io.Serializable;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.LinkedList;
import java.util.List;

import rice.p2p.commonapi.Id;

public class GroupState implements Serializable {
	
	private static final int MAX_LIST_LENGTH=5;

	public enum State {BOOTSTRAPPING, ACTIVE};
	
	private final InetAddress group;
	
	private final List<InetSocketAddress> activePeers=new LinkedList<InetSocketAddress>();
	
	private State currentState=State.BOOTSTRAPPING;

	private final Id bootstrappingNode;
	
	public GroupState(InetAddress group, Id bootstrappingNode) {
		this.group = group;
		this.bootstrappingNode = bootstrappingNode;		
	}

	public List<InetSocketAddress> getActivePeers() {
		return activePeers;
	}

	public Id getBootstrappingNode() {
		return bootstrappingNode;
	}

	public State getCurrentState() {
		return currentState;
	}
	
	public synchronized void addActivePeer(InetSocketAddress activePeer) {
		
		if (currentState==State.BOOTSTRAPPING) {
			currentState=State.ACTIVE;
		}
		
		activePeers.remove(activePeer);
		
		if (activePeers.size() > MAX_LIST_LENGTH) {
			activePeers.remove(0);
		}
		activePeers.add(activePeer);
	}

	public InetAddress getGroup() {
		return group;
	}

	public synchronized boolean removeActivePeer(InetSocketAddress endpoint) {
		activePeers.remove(endpoint);
		return activePeers.isEmpty();
	}
	

	
}
