package ch.unibe.iam.mcast.router.pastry;

import java.net.InetAddress;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.router.OverlayNetwork;
import ch.unibe.iam.mcast.router.OverlayNetworkListener;
import ch.unibe.iam.mcast.router.pastry.scribe.ScribeRing;

public class PastryOverlayNetwork implements OverlayNetwork {
	
	private static final int FIRST_PORT=31337;
	private static final int LAST_PORT=FIRST_PORT+2000;
	private static final Log log=LogFactory.getLog(PastryOverlayNetwork.class);
	
	private final OverlayNetworkListener listener;
	private final Map<InetAddress, ScribeRing> activeRings=new HashMap<InetAddress, ScribeRing>();
	private final Set<InetAddress> currentlyReserving=new HashSet<InetAddress>();
	private Set<Integer> unusedPorts=new HashSet<Integer>();

	public PastryOverlayNetwork(OverlayNetworkListener listener) {
		this.listener = listener;
		for (int i=FIRST_PORT; i<=LAST_PORT; i++) {
			unusedPorts.add(i);
		}
	}

	public void joinGroup(InetAddress group) {
		synchronized (this) {
			if (currentlyReserving.contains(group) || activeRings.containsKey(group)) {
				return;
			} else {
				currentlyReserving.add(group);
			}
		}
		
		ScribeRing ring=createPeer(group);
		
		
		synchronized(this) {
			currentlyReserving.remove(group);
			activeRings.put(group, ring);
		}
		
	}
	
	private ScribeRing createPeer(InetAddress group) {
		return null;
	}
	
	public synchronized int reservePort() {
		if (unusedPorts.isEmpty()) {
			throw new RuntimeException("all ports used!");
		}
		Iterator<Integer> it=unusedPorts.iterator();
		int ret=it.next();
		it.remove();
		return ret;
	}

	public synchronized void freePort(int port) {
		if (unusedPorts.contains(port)) {
			log.warn("trying to free unused port "+port);
		} else {
			unusedPorts.add(port);
		}
	}
	
	

	public void leaveGroup(InetAddress group) {
		// TODO Auto-generated method stub
		
	}
	
	

}
