package ch.unibe.iam.mcast.router.pastry;

import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.unibe.iam.mcast.router.pastry.message.BackOffMessage;
import ch.unibe.iam.mcast.router.pastry.message.BootstrapMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinToMessage;
import ch.unibe.iam.mcast.router.pastry.message.JoinedMessage;
import ch.unibe.iam.mcast.router.pastry.message.WantToJoinMessage;

import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.Message;

public class PeerState {
	private static final long BACKOFF_TIMEOUT=6000;
	private static final Log log=LogFactory.getLog(PeerState.class);
	
	private final Map<InetAddress, GroupState> groupStates = new HashMap<InetAddress, GroupState>();

	private final PeerContext ctx;

	public PeerState(PeerContext ctx) {
		this.ctx = ctx;

	}

	public synchronized void processMessage(Id from, Message message) {
		if (message instanceof WantToJoinMessage) {
			WantToJoinMessage msg = (WantToJoinMessage) message;
			if (ctx.isMasterFor(msg.getGroup())) {
				GroupState gs = groupStates.get(msg.getGroup());
				if (gs == null) {
					gs = new GroupState(msg.getGroup(), from);
					groupStates.put(msg.getGroup(), gs);
					ctx.sendMessage(from, new BootstrapMessage(msg.getGroup()));
				} else {
					switch (gs.getCurrentState()) {
						case BOOTSTRAPPING:
							ctx.sendMessage(from, new BackOffMessage(msg.getGroup(),BACKOFF_TIMEOUT));
							break;
						case ACTIVE:
							ctx.sendMessage(from, new JoinToMessage(msg.getGroup(),gs.getActivePeers()));
							break;
							default:
								log.warn("unknown state "+gs.getCurrentState());

					}
				}
			} 
		} else if (message instanceof JoinedMessage) {
			JoinedMessage msg=(JoinedMessage) message;
			if (ctx.isMasterFor(msg.getGroup())) {
				GroupState gs = groupStates.get(msg.getGroup());
				if (gs == null) {
					gs = new GroupState(msg.getGroup(), from);
					groupStates.put(msg.getGroup(), gs);
					ctx.sendMessage(from, new BackOffMessage(msg.getGroup(),BACKOFF_TIMEOUT));
				} else {
					switch (gs.getCurrentState()) {
						case BOOTSTRAPPING:
							ctx.sendMessage(from, new BackOffMessage(msg.getGroup(),BACKOFF_TIMEOUT));
							break;
						case ACTIVE:
							ctx.sendMessage(from, new JoinToMessage(msg.getGroup(),gs.getActivePeers()));
							break;
							default:
								log.warn("unknown state "+gs.getCurrentState());

					}
				}
				
			}
		}

	}

}
