package ch.unibe.iam.mcast.router.pastry.past;

import java.io.Serializable;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import rice.p2p.commonapi.Id;
import rice.p2p.past.PastContent;
import rice.p2p.past.PastException;

public class ActiveRingContent extends GroupBootstrapContent {

	private static final long ENTRY_EXPIRATION_TIMEOUT = 20000;

	private static final int MAX_SIZE = 20;

	private final Id firstNodeId;

	private final ArrayList<BootNodeEntry> activeEntries = new ArrayList<BootNodeEntry>();

	public ActiveRingContent(Id id, Id firstNodeId) {
		super(id);
		this.firstNodeId = firstNodeId;
	}

	public PastContent checkInsertHook(Id id, PastContent existingContent) throws PastException {

		// System.out.println("active ring content check insert");

		if (existingContent instanceof CreateNewRingContent) {
			CreateNewRingContent cnrc = (CreateNewRingContent) existingContent;
			if (cnrc.getBootstrapId().equals(firstNodeId)) {
				return this;
			} else {
				return existingContent;
			}

		}

		return existingContent;

	}

	public Id getFirstNodeId() {
		return firstNodeId;
	}

	public synchronized List<InetSocketAddress> getActiveEntries() {

		ArrayList<InetSocketAddress> ret = new ArrayList<InetSocketAddress>();
		for (Iterator<BootNodeEntry> i = activeEntries.iterator(); i.hasNext();) {
			BootNodeEntry bne = i.next();
			if (bne.isExpired()) {
				i.remove();
			} else {
				ret.add(bne.getAddress());
			}
		}

		return ret;

	}

	public synchronized List<InetSocketAddress> getBootstrapEntries() {

		ArrayList<InetSocketAddress> ret = new ArrayList<InetSocketAddress>();
		for (Iterator<BootNodeEntry> i = activeEntries.iterator(); i.hasNext();) {
			BootNodeEntry bne = i.next();
			if (bne.isExpired()) {
				i.remove();
			} else {
				if (!isLocalAddress(bne.getAddress().getAddress()))
				ret.add(bne.getAddress());
			}
		}

		return ret;

	}
	
	private boolean isLocalAddress(InetAddress addr) {
		try {
			return NetworkInterface.getByInetAddress(addr) != null;
		} catch (SocketException e) {
			throw new RuntimeException("this should never happend",e);
		}
	}

	public synchronized void refresh(InetSocketAddress address) {

		for (Iterator<BootNodeEntry> it = activeEntries.iterator(); it.hasNext();) {
			BootNodeEntry bne = it.next();
			if (bne.getAddress().equals(address)) {
				it.remove();
				bne.touch();
				activeEntries.add(0, bne);
				return;
			} else if (bne.isExpired()) {
				it.remove();
			}
		}

		if (activeEntries.size() >= MAX_SIZE) {
			activeEntries.remove(activeEntries.size() - 1);
		}

		activeEntries.add(new BootNodeEntry(address));

	}

	public synchronized void remove(InetSocketAddress endpoint) {

		for (Iterator<BootNodeEntry> it = activeEntries.iterator(); it.hasNext();) {
			BootNodeEntry bne = it.next();
			if (bne.getAddress().equals(endpoint)) {
				it.remove();
			}
		}

	}

	public String toString() {
		return "ActiveRing" + activeEntries;
	}

	private static class BootNodeEntry implements Serializable {

		private final InetSocketAddress address;

		private long expirationTime;

		public InetSocketAddress getAddress() {
			return address;
		}

		public BootNodeEntry(InetSocketAddress address) {
			this.address = address;
			touch();
		}

		private void touch() {
			this.expirationTime = System.currentTimeMillis() + ENTRY_EXPIRATION_TIMEOUT;
		}

		public boolean isExpired() {
			return System.currentTimeMillis() > expirationTime;
		}

		public String toString() {
			return address.getAddress().getHostAddress() + "/" + (new Date(expirationTime).toString());
		}

	}

}
