package ch.unibe.iam.mcast.router.pastry.scribe;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rice.Destructable;
import rice.environment.Environment;
import rice.p2p.commonapi.Application;
import rice.p2p.commonapi.Endpoint;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.Message;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.RouteMessage;
import rice.p2p.scribe.ScribeClient;
import rice.p2p.scribe.ScribeContent;
import rice.p2p.scribe.ScribeImpl;
import rice.p2p.scribe.Topic;
import rice.pastry.NodeIdFactory;
import rice.pastry.socket.SocketNodeHandle;
import rice.pastry.socket.SocketPastryNode;
import rice.pastry.socket.SocketPastryNodeFactory;
import rice.pastry.standard.RandomNodeIdFactory;
import ch.unibe.iam.mcast.router.OverlayTopologyChangeListener;
import ch.unibe.iam.mcast.router.pastry.message.DataAddressQuery;
import ch.unibe.iam.mcast.router.pastry.message.DataAddressResponse;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;

public class ScribeRing implements ScribeClient, Application, Destructable {

	private final Environment environment = new Environment();

	private final Log log = LogFactory.getLog(ScribeRing.class);

	private final NodeIdFactory fact = new RandomNodeIdFactory(environment);

	private final Endpoint endpoint;

	private final ScribeImpl scribe;

	private final InetSocketAddress dataSocketAddress;

	// private final Map<Id, InetSocketAddress> childrenDataAddresses = new
	// HashMap<Id, InetSocketAddress>();
	private final Map<Id, InetSocketAddress> parentDataAddresses = new HashMap<Id, InetSocketAddress>();

	private final OverlayTopologyChangeListener listener;

	private final InetAddress groupAddress;

	private final Id myId;

	private final InetSocketAddress localAddress;

	// private final SocketPastryNodeFactory factory;

	private final SocketPastryNode node;

	private final Topic topic = new Topic(rice.pastry.Id.build(new int[] { -1, -1, -1, -1, -1 }));

	private final Random random = new Random();

	private ParentWatch parentWatch;

	public ScribeRing(QoSRequirement qosRequrement, InetSocketAddress[] addresses, SocketPastryNodeFactory nodeFactory, InetSocketAddress dataSocketAddress,
			InetAddress groupAddress, OverlayTopologyChangeListener listener) throws IOException, InterruptedException {

		
		environment.getParameters().setBoolean("pastry_factory_selectorPerNode", true); 		
		environment.getParameters().setBoolean("pastry_factory_processorPerNode", true);												
		environment.getParameters().setBoolean("pastry_factory_multipleNodes", true); 
		environment.getParameters().setInt("pastry_factory_bootsInParallel", 5);

		
		// factory = localAddress.getAddress() == null ? new
		// SocketPastryNodeFactory(fact, localAddress.getPort(), environment) :
		// new SocketPastryNodeFactory(fact, localAddress
		// .getAddress(), localAddress.getPort(), environment);
		this.myId = createId(qosRequrement);
		ArrayList<InetSocketAddress> bs=new ArrayList<InetSocketAddress>();
		if (addresses != null)
			Collections.addAll(bs, addresses);
		System.out.println("bootstrapping group "+groupAddress+" from "+bs);

		rice.pastry.NodeHandle bootstrapNode = addresses == null ? null : nodeFactory.getNodeHandle(addresses);

		node = (SocketPastryNode) nodeFactory.newNode(bootstrapNode, (rice.pastry.Id) myId);

		while (!node.isReady()) {
			Thread.sleep(100);
		}

		this.localAddress = ((SocketNodeHandle) node.getLocalNodeHandle()).getAddress();
		System.out.println("peer: " + localAddress + " ready");
		this.dataSocketAddress = dataSocketAddress;
		this.groupAddress = groupAddress;
		this.listener = listener;

		this.endpoint = node.buildEndpoint(this, "myinstance");
		this.endpoint.register();

		scribe = new ScribeImpl(node, "ring");
		scribe.subscribe(topic, this);
		this.parentWatch = new ParentWatch();
		this.parentWatch.start();
	}

	private int[] parsePreDefinedID() {
		
		if (Boolean.parseBoolean(System.getenv("FORCE_ROOT"))) {
			return new int[]{-1, -1, -1, -1, -1};
		}
			
		
		String val = System.getenv("PEER_ID");		
		
		if (val == null) {
			return null;
		} else {
			String[] vals = val.split("\\s+");
			int ret[] = new int[vals.length];
			for (int i = 0; i < vals.length; i++) {
				ret[i] = (int) (Long.parseLong(vals[i], 16) & 0xffffffff);
			}
			return ret;
		}
	}

	public Id createId(QoSRequirement requirement) {
		int len = rice.pastry.Id.nlen;
		int material[] = parsePreDefinedID();
		if (material == null) {
			material = new int[len];

			for (int i = 0; i < len - 1; i++) {
				material[i] = random.nextInt();
			}
			
			
			
			long requestedBandwidth = requirement == null ? 0 : requirement.getBandwidth();
			
			
			
			int bw=requestedBandwidth > (long)Integer.MAX_VALUE ? Integer.MAX_VALUE : (int)requestedBandwidth;
			
			int firstDigit = bw;
//			if (requirement != null) {
//				switch (requirement) {
//				case HIGH:
//					firstDigit = 0xff000000;
//					break;
//				case MEDIUM:
//					firstDigit = 0x7f000000;
//					break;
//				default:
//					firstDigit = 0;
//				}
//			}

			material[len - 1] = firstDigit;

			// material[len-1]=
			// NodeId generated=NodeId.buildNodeId().buildNodeId()
		}
		return rice.pastry.Id.build(material);
	}

	public boolean anycast(Topic topic, ScribeContent content) {
		return false;
	}

	public void childAdded(Topic topic, NodeHandle child) {
		// System.out.println("child added " + child + " for peer: " +
		// localAddress);
		// DataAddressQuery query = null;
		// synchronized (childrenDataAddresses) {
		// if (!childrenDataAddresses.containsKey(child.getId())) {
		// childrenDataAddresses.put(child.getId(), null);
		// query = new DataAddressQuery(groupAddress, myId);
		//
		// } else {
		// log.warn("new child is already existing: " +
		// child.getId().toStringFull());
		// }
		// }
		//
		// if (query != null) {
		// endpoint.route(null, query, child);
		// }
	}

	// public void onNewParent(NodeHandle parent) {
	// // System.out.println("child added " + child + " for peer: " +
	// localAddress);
	// DataAddressQuery query = null;
	// synchronized (childrenDataAddresses) {
	// if (!childrenDataAddresses.containsKey(child.getId())) {
	// childrenDataAddresses.put(child.getId(), null);
	// query = new DataAddressQuery(groupAddress, myId);
	//
	// } else {
	// log.warn("new child is already existing: " +
	// child.getId().toStringFull());
	// }
	// }
	//
	// if (query != null) {
	// endpoint.route(null, query, child);
	// }
	// }

	public void childRemoved(Topic topic, NodeHandle child) {
		// InetSocketAddress childAddress;
		// synchronized (childrenDataAddresses) {
		// childAddress = childrenDataAddresses.get(child.getId());
		// }
		//
		// if (childAddress != null) {
		// listener.removeConnection(groupAddress, childAddress);
		// }
	}

	public void deliver(Topic topic, ScribeContent content) {

	}

	public void subscribeFailed(Topic topic) {
		log.fatal("could not join the topic " + topic.getId().toStringFull() + " peer: " + this.localAddress);
		scribe.subscribe(topic, this);
	}

	public void deliver(Id id, Message message) {
		if (message instanceof DataAddressQuery) {
			DataAddressQuery query = (DataAddressQuery) message;
			if (query.getGroupAddress().equals(this.groupAddress)) {
				System.out.println("== seding response with data address " + dataSocketAddress);
				DataAddressResponse response = new DataAddressResponse(myId, groupAddress, dataSocketAddress);
				endpoint.route(query.getSourceAddress(), response, null);
			} else {
				log.warn("received address query " + query + " for some other group than " + groupAddress.getHostAddress());
			}

		} else if (message instanceof DataAddressResponse) {
			DataAddressResponse response = (DataAddressResponse) message;
			InetSocketAddress dataAddress = null;
			if (response.getGroupAddres().equals(groupAddress)) {
				synchronized (parentDataAddresses) {
					if (!parentDataAddresses.containsKey(response.getSource())) {
						log.warn("received DataAddressRespone from unknown child " + response.getSource());
					} else {
						if (parentDataAddresses.get(response.getSource()) == null) {
							parentDataAddresses.put(response.getSource(), response.getDataAddress());
							dataAddress = response.getDataAddress();
						} else {
							log.warn("received data adress response from already known child " + response);
						}
					}
				}
				if (dataAddress != null) {
					listener.addConnection(groupAddress, dataAddress);
				}
			} else {
				log.warn("received address response " + response + " for some other group than " + groupAddress.getHostAddress());
			}
		} else {
			log.warn("received unknown message " + message);
		}

	}

	public boolean forward(RouteMessage message) {
		System.out.println("peer: " + localAddress + " forwarding " + message);
		return true;
	}

	public void update(NodeHandle handle, boolean joined) {
	}

	public void stop() {
		scribe.unsubscribe(topic, this);
		node.destroy();
		node.getEnvironment().destroy();
		parentWatch.interrupt();
		// environment.destroy();
		// scribe.unsubscribe(topic, this);
		// node.destroy();
		// factory.
		// environment.destroy();
	}

	public InetSocketAddress getLocalAddress() {
		return localAddress;
	}

	public static void main(String argv[]) throws UnknownHostException, IOException, InterruptedException {

//		Environment environment = new Environment();
//
//		environment.getParameters().setBoolean("pastry_factory_selectorPerNode", true); // set
//		// this
//		// before
//		// you
//		// create
//		// the
//		// PastryNodeFactory
//		environment.getParameters().setBoolean("pastry_factory_processorPerNode", true); // set
//		// this
//		// before
//		// you
//		// create
//		// the
//		// PastryNodeFactory
//		environment.getParameters().setBoolean("pastry_factory_multipleNodes", true); // set
//		// this
//		// before
//		// you
//		// create
//		// the
//		// PastryNodeFactory
//
//		SocketPastryNodeFactory nodeFactory = new SocketPastryNodeFactory(new RandomNodeIdFactory(environment), 1234, environment);
//
//		InetSocketAddress ring1Address = new InetSocketAddress("localhost", 1234);
//		// InetSocketAddress ring2Address = new InetSocketAddress("localhost",
//		// 1235);
//		// InetSocketAddress ring3Address = new InetSocketAddress("localhost",
//		// 1236);
//		InetAddress groupAddress = InetAddress.getByAddress(new byte[] { (byte) 224, 0, 0, 1 });
//		InetSocketAddress dataPort1 = new InetSocketAddress("localhost", 2222);
//		InetSocketAddress dataPort2 = new InetSocketAddress("localhost", 2223);
//		InetSocketAddress dataPort3 = new InetSocketAddress("localhost", 2224);
//		ScribeRing r1 = new ScribeRing(QoSRequirement.BEST_EFFORT, null, nodeFactory, dataPort1, groupAddress, new TestListener("ring 1"));
//		Thread.sleep(1000);
//		System.out.println("starting ring 2 ...");
//		ScribeRing r2 = new ScribeRing(QoSRequirement.BEST_EFFORT, new InetSocketAddress[] { r1.getLocalAddress() }, nodeFactory, dataPort2, groupAddress, new TestListener(
//				"ring 2"));
//		Thread.sleep(1000);
//		System.out.println("starting ring 3 ...");
//		ScribeRing r3 = new ScribeRing(QoSRequirement.BEST_EFFORT, new InetSocketAddress[] { r1.getLocalAddress(), r2.getLocalAddress() }, nodeFactory, dataPort3, groupAddress,
//				new TestListener("ring 3"));
//		Thread.sleep(20000);
//		System.out.println("stopping ring 2");
//		r2.stop();
//		Thread.sleep(1000);
//		System.out.println("starting ring 4 ...");
//		ScribeRing r4 = new ScribeRing(QoSRequirement.BEST_EFFORT, new InetSocketAddress[] { r1.getLocalAddress(), r2.getLocalAddress() }, nodeFactory, dataPort2, groupAddress,
//				new TestListener("ring 4"));
//		System.out.println("stopped ring 2");
//		Thread.sleep(2000000);
	}

	private static class TestListener implements OverlayTopologyChangeListener {

		private final String name;

		public TestListener(String name) {
			this.name = name;

		}

		public void addConnection(InetAddress group, InetSocketAddress connection) {
			System.out.println("adding connection for group " + group + " to " + connection + " for " + name);

		}

		public void removeConnection(InetAddress group, InetSocketAddress connection) {
			System.out.println("removing connection for group " + group + " to " + connection + " for " + name);
		}

	}

	public void destroy() {
		System.out.println("scribe ring " + localAddress + " destroyed ");

	}

	private class ParentWatch extends Thread {
		private NodeHandle parentHandle = null;

		@Override
		public void run() {
			while (!interrupted()) {
				NodeHandle currentParent = scribe.getParent(topic);
				if (currentParent != parentHandle) {
					if (parentHandle != null) {
						InetSocketAddress remoteAddress;
						synchronized (parentDataAddresses) {
							remoteAddress = parentDataAddresses.remove(parentHandle.getId());
						}
						if (remoteAddress != null) {
							listener.removeConnection(groupAddress, remoteAddress);
						} else {
							log.warn("could not find data port for parent " + parentHandle.getId());
						}
					}

					if (currentParent != null) {
						DataAddressQuery query = null;
						synchronized (parentDataAddresses) {
							if (!parentDataAddresses.containsKey(currentParent.getId())) {
								parentDataAddresses.put(currentParent.getId(), null);
								query = new DataAddressQuery(groupAddress, myId);

							} else {
								log.warn("new child is already existing: " + currentParent.getId().toStringFull());
							}
						}

						if (query != null) {
							endpoint.route(null, query, currentParent);
						}
					}
					parentHandle = currentParent;
				}

				try {
					sleep(500);
				} catch (InterruptedException e) {
					return;
				}
			}
		}
	}

}
