package ch.unibe.iam.mcast.webapp.ws.euqos;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import ch.unibe.iam.mcast.hivemind.component.iface.MulticastRouter;
import ch.unibe.iam.mcast.router.qos.QoSRequirement;

public class EuQoSReservationService implements ServletContextListener {

	private static int counter;

	private static MulticastRouter router;

	private static Map<String, InetAddress> sessionGroups = new HashMap<String, InetAddress>();

	public String session(String destinationAddress, int destinationPort, int protocolId) {

		String id;
		synchronized (EuQoSReservationService.class) {
			id = "euqos-mm-" + (counter++);
			try {
				sessionGroups.put(id, InetAddress.getByName(destinationAddress));
			} catch (UnknownHostException e) {
				throw new RuntimeException(e);
			}
		}

		System.out.println("starting session " + id);
		return id;
	}

	public boolean reserve(String sessionId, int bandwidth, int max_delay, int jitter) {
		InetAddress group;
		synchronized (EuQoSReservationService.class) {
			group = sessionGroups.get(sessionId);
		}

		System.out.println("group: " + group);
		if (group != null) {
			System.out.println("changing ...");
			router.changeQoSReservation(group, new QoSRequirement(bandwidth, max_delay, jitter));
		}
		return true;
	}

	public void release(String sessionId) {
		InetAddress group;
		synchronized (EuQoSReservationService.class) {
			group = sessionGroups.remove(sessionId);
		}
		
		if (group != null) {
			router.changeQoSReservation(group, new QoSRequirement(0, 0, 0));
		}
	}

	public synchronized void refresh(String sessionId) {
		// System.out.println("refreshing " + sessionId);
	}

	static synchronized void setRouter(MulticastRouter router) {
		EuQoSReservationService.router = router;
	}

	public void contextDestroyed(ServletContextEvent evt) {
		// TODO Auto-generated method stub

	}

	public void contextInitialized(ServletContextEvent evt) {
		setRouter((MulticastRouter) evt.getServletContext().getAttribute("multicast.router"));

	}
}
